__all__ = ()

from time import perf_counter

from ...backend.futures import sleep, Task, future_or_timeout, Future, CancelledError, skip_ready_cycle

from ..core import KOKORO

HEARTBEAT_TIMEOUT = 20.0


class Kokoro:
    """
    A beater for Discord gateways.
    
    Attributes
    ----------
    beat_task : `None` or ``Task``
        The beat task of kokoro's gateway. Set by ``._task`` meanwhile it's gateway is connected.
    beat_waiter : `None` or ``Future``
        `.beater` task's waiter. Sleeps `.beater` between two beat.
    beater : `None` or ``Task`` of ``._keep_beating``
        A task what sends a beating payload to Discord. Set only meanwhile kokoro is in the middle of a beat.
    gateway : ``DiscordGateway`` or ``DiscordGatewayVoice``
        Kokoro's owner gateway.
    interval : `float`
        The time needed to be waited between two beats.
    last_answer : `float`
        The time of the last answered heartbeat event in `perf_counter` time.
    last_send : `float`
        The last time when kokoro sent a heartbeat event in `last_send` time.
    latency : `float`
        The time interval between the last beat and the response's time. Defaults to `.DEFAULT_LATENCY` if no beat was
        done yet.
    running : `bool`
        Whether the kokoro should run and restart itself.
    should_beat : `bool`
        Whether the kokoro should keep beating.
    task : `None` or ``Task`` of ``._start``
        The main keep alive task of kokoro.
    ws_waiter : `None` or ``Future``
        The waiter of kokoro, what waits for it's gateway to connect it's websocket.
    
    Class Attributes
    ----------------
    DEFAULT_LATENCY : `float` = `9999.0`
        The default latency of kokoro. Better than using `inf`.
    """
    __slots__ = ('beat_task', 'beat_waiter', 'beater', 'gateway', 'interval', 'last_answer', 'last_send', 'latency',
        'running', 'should_beat', 'task', 'ws_waiter')
    
    DEFAULT_LATENCY = 9999.0
    
    async def __new__(cls, gateway):
        """
        Creates a ``Kokoro`` instance bound to it's gateway.
        
        This method is a coroutine.
        
        Parameters
        ----------
        gateway : ``DiscordGateway`` or ``DiscordGatewayVoice``
            The gateway of kokoro.
        
        Returns
        -------
        kokoro : ``Kokoro``
        
        Notes
        -----
        Before the coroutine returns a full loop is done to ensure `._start` is already started up.
        """
        self = object.__new__(cls)
        
        self.gateway = gateway
        self.interval = 40.0 # we set it from outside
        self.running = False
        now_ = perf_counter()
        self.last_answer = now_
        self.last_send = now_
        self.latency = self.DEFAULT_LATENCY
        self.ws_waiter = None
        self.beater = None
        self.beat_task = None
        self.beat_waiter = None
        self.task = Task(self._start(), KOKORO)
        
        #skip 1 loop
        await skip_ready_cycle()
        
        return self
    
    
    async def restart(self):
        """
        Restarts kokoro.
        
        This method is a coroutine.
        """
        self.cancel()
        # skip 1 loop
        await skip_ready_cycle()
        self.task = Task(self._start(), KOKORO)
        # skip 1 loop
        await skip_ready_cycle()
    
    
    async def _start(self):
        """
        The main keep alive task of kokoro.
        
        The control flow of the method is the following:
            - Set `.ws_waiter` and wait till it is done. If it is done, means that kokoro's gateway is connected.
                If it was cancelled meanwhile, means the beat task should stop.
            - Sets `.beater` to ``._keep_beating`` and awaits it. This is the main beater of kokoro and runs meanwhile
                it's gateway is connected. This task can be cancelled, but we ignore that case.
            - If `.running` is still `True`, repeat.
        
        This method is a coroutine.
        """
        try:
            self.running = True
            while True:
                #wait for start
                try:
                    waiter = Future(KOKORO)
                    self.ws_waiter = waiter
                    await waiter
                except CancelledError:
                    #kokoro cancelled, client shuts down
                    break
                finally:
                    self.ws_waiter = None
                
                #keep beating
                try:
                    beater = Task(self._keep_beating(), KOKORO)
                    self.beater = beater
                    await beater
                except CancelledError:
                    #connection cancelled, lets wait for it
                    pass
                finally:
                    #make sure
                    self.beater = None
                
                if self.running:
                    continue
                
                break
        finally:
            if self.task is KOKORO.current_task:
                self.task = None
    
    
    async def _keep_beating(self):
        """
        The beater task of kokoro.
        
        The control flow of tze method is the following:
            - If `.should_beat` is not `True`, break out.
            - Wait `.interval` time and set the waiting ``Future`` to `self.beat_waiter`. If kokoro is cancelled, or
                if it should beat now, it is cancelled and we repeat the loop.
            - If we did not get answer since last beating (what is triggered first from outside), then we stop the
                gateway. We also break out from the loop to terminate the beating state and we will wait for the
                websocket to connect again.
            - We beat one with starting `gateway._beat` as a ``Task`` and setting it to `.beat_task`. If the task is
                not completed before it's respective timeout, we stop the gateway here as well. We also break out
                from the loop to terminate the beating state and we will wait for the websocket to connect again.
                This task can also be cancelled. If cancellation occurs, we repeat the loop.
            - If the beating task is done, we update `.last_send` to the current `perf_counter` time. Repeat the loop.
        
        This method is a coroutine.
        """
        self.last_answer = perf_counter()
        gateway = self.gateway
        self.should_beat = True
        while self.should_beat:
            waiter = sleep(self.interval, KOKORO)
            self.beat_waiter = waiter
            try:
                await waiter
            except CancelledError:
                self.last_send = perf_counter()
                continue
            finally:
                self.beat_waiter = None
            
            if (self.last_answer+self.interval+HEARTBEAT_TIMEOUT)-perf_counter() <= 0.0:
                self.should_beat = False
                Task(gateway.terminate(), KOKORO)
                break
            
            try:
                task = Task(gateway._beat(), KOKORO)
                future_or_timeout(task, HEARTBEAT_TIMEOUT)
                self.beat_task = task
                await task
            except TimeoutError:
                self.should_beat = False
                Task(gateway.terminate(), KOKORO)
                break
            except CancelledError:
                continue
            finally:
                self.beat_task = None
            
            self.last_send = perf_counter()
    
    
    def start_beating(self):
        """
        Starts kokoro's beating. Handles all the cases around the board to do so.
        """
        # case 1 : we are not running
        if not self.running:
            Task(self._start_beating(), KOKORO)
            return
        
        #case 2 : we wait for ws
        waiter = self.ws_waiter
        if (waiter is not None):
            self.ws_waiter = None
            waiter.set_result(None)
            return
        
        #case 3 : we wait for beat response
        waiter = self.beat_waiter
        if (waiter is not None):
            self.beat_waiter = None
            waiter.cancel()
            return
        
        #case 4 : we are beating
        task = self.beat_task
        if (task is not None):
            self.beat_waiter = None
            task.cancel()
            return
    
    
    async def _start_beating(self):
        """
        Internal method to start beating when kokoro is not running.
        
        This method is a coroutine.
        """
        # starts kokoro, then beating
        self.task = Task(self._start(), KOKORO)
        # skip 1 loop
        await skip_ready_cycle()
        
        waiter = self.ws_waiter
        if (waiter is not None):
            self.ws_waiter = None
            waiter.set_result(None)
    
    
    def answered(self):
        """
        When gateway received answer after beating, it's gateway calls this method.
        
        This method updates `.last_answer` and `.latency` as well. Because of the updated `.last_answer`,
        ``._keep_beating`` task will know that it's gateway is still connected.
        """
        now_ = perf_counter()
        self.last_answer = now_
        self.latency = now_-self.last_send
    
    
    def terminate(self):
        """
        Terminates kokoro till restart. Called when it's gateway will resume it's connection.
        """
        #case 1 : we are not running
        if not self.running:
            return
        
        #case 2 : we are waiting for ws
        waiter = self.ws_waiter
        if (waiter is not None):
            # it is fine, it is what we should do
            return
        
        #case 2 : we are beating
        beater = self.beater
        if (beater is not None):
            self.should_beat = False
            
            #case 3.1 : we wait to beat
            waiter = self.beat_waiter
            if (waiter is not None):
                self.beat_waiter = None
                waiter.cancel()
                return
            
            #case 3.2: we are beating
            task = self.beat_task
            if (task is not None):
                self.beat_task = None
                task.cancel()
                return
    
    
    def cancel(self):
        """
        Stops kokoro. Called when it's gateway is closing.
        """
        #case 1 : we are not running
        if not self.running:
            return
        self.running = False
        
        #case 2 : we are waiting for ws
        waiter = self.ws_waiter
        if (waiter is not None):
            self.ws_waiter = None
            waiter.cancel()
            return
        
        #case 3 : we are beating
        beater = self.beater
        if beater is not None:
            self.should_beat = False
            
            #case 3.1 : we wait to beat
            waiter = self.beat_waiter
            if (waiter is not None):
                self.beat_waiter = None
                waiter.cancel()
                return
            
            #case 3.2: we are beating
            task = self.beat_task
            if (task is not None):
                self.beat_task = None
                task.cancel()
                return
    
    
    async def beat_now(self):
        """
        Sends a heartbeat event now. If kokoro is not at beating state, or anywhere in between, resets itself as it just
        received a heartbeat answer right now.
        
        This method is a coroutine.
        """
        while True: #GOTO
            #case 1 : we are not running:
            if not self.running:
                Task(self._start_beating(), KOKORO)
                should_beat_now = True # True = send beat data
                break
            
            # case 2 : we wait for ws
            waiter = self.ws_waiter
            if (waiter is not None):
                self.ws_waiter = None
                waiter.set_result(None)
                should_beat_now = True # True = send beat data
                break
            
            # case 3 : we wait to beat
            waiter = self.beat_waiter
            if (waiter is not None):
                # better skip a beat
                self.beat_waiter = None
                waiter.cancel()
                should_beat_now = True # True = send beat data
                break
            
            # case 4 : we already beat
#            task = self.beat_task
#            if (task is not None):
#                should_beat_now = False
#                break
            
            should_beat_now = False
            break
        
        if should_beat_now:
            await self.gateway._beat()
        else:
            self.answered()
    
    # Fixed?
    """
    if __debug__:
        def __del__(self):
            # bug?
            waiter = self.ws_waiter
            if (waiter is not None):
                waiter.__silence__()
            
            # despair?
            waiter = self.beat_waiter
            if (waiter is not None):
                waiter.__silence__()
    """
    
    def __repr__(self):
        """Returns kokoro's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' client=',
            repr(self.gateway.client.full_name),
            ' state=',
        ]
        
        if self.task is None:
            state = 'stopped'
        elif self.beater is None:
            state = 'waiting for websocket'
        else:
            state = 'beating'
        
        repr_parts.append(state)
        repr_parts.append('>')
        
        return ''.join(repr_parts)



__all__ = ('ComponentCommand', )

try:
    # CPython
    from re import Pattern
except ImportError:
    # ChadPython (PyPy)
    from re import _pattern_type as Pattern

from functools import partial as partial_func

from ...discord.interaction.components import COMPONENT_CUSTOM_ID_LENGTH_MAX
from ...discord.events.handling_helpers import route_value, Router, create_event_from_class
from .wrappers import SlasherCommandWrapper
from .utils import _check_maybe_route
from .converters import get_component_command_parameter_converters, RegexMatcher, \
    check_component_converters_satisfy_string, check_component_converters_satisfy_regex
from .responding import process_command_coroutine
from .exceptions import handle_command_exception, test_exception_handler, _register_exception_handler

SLASH_COMMAND_PARAMETER_NAMES = ('command', 'custom_id')

SLASH_COMMAND_NAME_NAME = None
SLASH_COMMAND_COMMAND_NAME = 'command'

def _validate_custom_id(custom_id):
    """
    Validates a `custom_id` value.
    
    Parameters
    ----------
    custom_id : `str` or `re.Pattern`
        The `custom_id` to validate.
    
    Returns
    -------
    custom_id : `str`or `re.Pattern`
        The validated custom_id.
    
    Raises
    ------
    ValueError
        If `custom_id`'s length is out of the expected range.
    """
    if isinstance(custom_id, str):
        if type(custom_id) is not str:
            custom_id = str(custom_id)
        
        custom_id_length = len(custom_id)
        if (custom_id_length < 1) or (custom_id_length > COMPONENT_CUSTOM_ID_LENGTH_MAX):
            raise ValueError(f'`custom_id` length can be in range [1:{COMPONENT_CUSTOM_ID_LENGTH_MAX}], got '
                f'{custom_id_length}; {custom_id!r}.')
    
    return custom_id


def _validate_custom_ids(custom_id):
    """
    Validates one or more `custom_id` values.
    
    Parameters
    ----------
    custom_id : `str`, (`list` or `set`) of `str`.
        The `custom_id` to validate.
    
    Returns
    -------
    custom_id : `set` of (`str` or `re.Pattern`)
        The non-duped custom-ids.
    
    Raises
    ------
    TypeError
        If `custom_id`'s type is incorrect.
    ValueError
        If a `custom_id`'s length is out of the expected range.
    """
    custom_ids = set()
    if isinstance(custom_id, (str, Pattern)):
        custom_id = _validate_custom_id(custom_id)
        custom_ids.add(custom_id)
    elif isinstance(custom_id, (list, set)):
        
        
        for sub_custom_id in custom_id:
            if isinstance(sub_custom_id, (str, Pattern)):
                sub_custom_id = _validate_custom_id(sub_custom_id)
                custom_ids.add(sub_custom_id)
                continue
            
            raise TypeError(f'`custom_id` contains a non `st` element, got: {custom_id.__class__.__name__}.')
        
        if not custom_id:
            raise ValueError(f'`custom_id` received as empty {custom_id.__class__.__name__}.')
    
    else:
        raise TypeError(f'`custom_id` can be given as `str`, (`list`, `set`) of `str`, got '
            f'{custom_id.__class__.__name__}.')
    
    return custom_ids


def split_and_check_satisfaction(custom_ids, parameter_converters):
    """
    Splits custom id-s to `str` and to `re.Pattern`-s and validates them.
    
    Parameters
    ----------
    custom_ids : `set` of (`str` or `re.Pattern`)
        The custom-ids to split and validate.
    parameter_converters : `tuple` of ``ParameterConverter``
        The parameter converters generated from a component command.
    
    Returns
    -------
    string_custom_ids : `None` or `tuple` of `str`
        String custom ids.
    regex_custom_ids : `None` or `tuple` of ``RegexMatcher``
        Regex custom ids.
    
    Raises
    ------
    ValueError
        A string or regex pattern is not satisfied.
    """
    # Build
    string_custom_ids = None
    regex_custom_ids = None
    for custom_id in custom_ids:
        if isinstance(custom_id, str):
            if string_custom_ids is None:
                string_custom_ids = []
            
            string_custom_ids.append(custom_id)
        else:
            if regex_custom_ids is None:
                regex_custom_ids = []
            
            regex_custom_ids.append(custom_id)
    
    # Convert
    if (string_custom_ids is not None):
        string_custom_ids = tuple(string_custom_ids)
    
    if (regex_custom_ids is not None):
        regex_custom_ids = tuple(RegexMatcher(regex_custom_id) for regex_custom_id in regex_custom_ids)
    
    # Check
    if (string_custom_ids is not None):
        check_component_converters_satisfy_string(parameter_converters)
    
    if (regex_custom_ids is not None):
        for regex_custom_id in regex_custom_ids:
            check_component_converters_satisfy_regex(parameter_converters, regex_custom_id)
    
    # Good
    return string_custom_ids, regex_custom_ids


class ComponentCommand:
    """
    A command, which is called if a command interaction is executed with a specific `custom_id`.
    
    Attributes
    ----------
    _command_function : `async-callable˛
        The command's function to call.
    _exception_handlers : `None` or `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
        
        Same as ``Slasher._exception_handlers``.
    
    _parent_reference : `None` or ``WeakReferer`` to ``SlasherApplicationCommand``
        The parent slasher of the component command.
    _parameter_converters : `tuple` of ``ParameterConverter``
        Parsers to parse command parameters.
    _string_custom_ids : `None` or `tuple` of `str`
        The custom id-s to wait for.
    _regex_custom_ids : `None` or `tuple` of `re.Pattern`.
        Regex pattern to match custom-ids.
    """
    __slots__ = ('_command_function', '_exception_handlers', '_parent_reference', '_parameter_converters',
        '_regex_custom_ids', '_string_custom_ids')
    
    
    @classmethod
    def from_class(cls, klass):
        """
        Creates a new ``ComponentCommand`` instance from the given `klass`.
        
        Parameters
        ----------
        klass : `type`
            The class to create component command from.
        
        Returns
        -------
        self : ``ComponentCommand`` or ``Router``
        
        Raises
        ------
        TypeError
            If any attribute's type is incorrect.
        ValueError
            If any attribute's value is incorrect.
        """
        return create_event_from_class(cls, klass, SLASH_COMMAND_PARAMETER_NAMES, SLASH_COMMAND_NAME_NAME,
            SLASH_COMMAND_COMMAND_NAME)
    
    
    def __new__(cls, func, custom_id):
        """
        Creates a new ``ComponentCommand`` instance with the given parameters
        
        Parameters
        ----------
        func : `None` or `async-callable`, Optional
            The function used as the command when using the respective slash command.
        custom_id : `str`, (`list` or `set`) of `str`, `tuple` of (`str`, (`list` or `set`) of `str`)
            Custom id to match by the component command.
        
        Returns
        -------
        self : ``ComponentCommand`` or ``Router``
        
        Raises
        ------
        TypeError
            - If `func` is not async callable, neither cannot be instanced to async.
            - If `func` accepts keyword only parameters.
            - If `func` accepts `*args`.
            - If `func` accepts `**kwargs`
            - If `custom_id`'s type is incorrect.
        ValueError:
            - If no `custom_id` was received.
            - If `custom_id` contains incorrect value.
        """
        if (func is not None) and isinstance(func, SlasherCommandWrapper):
            command, wrappers = func.fetch_function_and_wrappers_back()
        else:
            command = func
            wrappers = None
        
        # Check for routing.
        route_to = 0
        custom_id, route_to = _check_maybe_route('name', custom_id, route_to, _validate_custom_ids)
        
        command, parameter_converters = get_component_command_parameter_converters(command)
        
        if route_to:
            custom_id = route_value(custom_id, route_to)
            
            router = []
            
            for custom_id in custom_id:
                string_custom_ids, regex_custom_ids = split_and_check_satisfaction(custom_id, parameter_converters)
                
                self = object.__new__(cls)
                self._command_function = command
                self._parameter_converters = parameter_converters
                self._string_custom_ids = string_custom_ids
                self._regex_custom_ids = regex_custom_ids
                self._parent_reference = None
                self._exception_handlers = None
                
                if (wrappers is not None):
                    for wrapper in wrappers:
                        wrapper.apply(self)
                
                router.append(self)
            
            return Router(router)
        else:
            string_custom_ids, regex_custom_ids = split_and_check_satisfaction(custom_id, parameter_converters)
            
            self = object.__new__(cls)
            self._command_function = command
            self._parameter_converters = parameter_converters
            self._string_custom_ids = string_custom_ids
            self._regex_custom_ids = regex_custom_ids
            self._parent_reference = None
            self._exception_handlers = None
            
            if (wrappers is not None):
                for wrapper in wrappers:
                    wrapper.apply(self)
            
            return self
    
    
    def __repr__(self):
        """Returns the component command's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        string_custom_ids = self._string_custom_ids
        if (string_custom_ids is None):
            field_added = False
        else:
            field_added = True
            
            repr_parts.append(' string_custom_ids=[')
            index = 0
            limit = len(string_custom_ids)
            
            while True:
                string_custom_id = string_custom_ids[index]
                repr_parts.append(repr(string_custom_id))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        regex_custom_ids = self._regex_custom_ids
        if (regex_custom_ids is not None):
            if field_added:
                repr_parts.append(', ')
            
            repr_parts.append(' regex_custom_ids=[')
            index = 0
            limit = len(regex_custom_ids)
            
            while True:
                regex_custom_id = regex_custom_ids[index]
                repr_parts.append(repr(regex_custom_id.pattern.pattern))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        return ''.join(repr_parts)
    
    
    async def __call__(self, client, interaction_event, regex_match):
        """
        Calls the component command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        regex_match : `None` or ``RegexMatch``
            The matched regex if applicable.
        """
        parameters = []
        
        for parameter_converter in self._parameter_converters:
            try:
                parameter = await parameter_converter(client, interaction_event, regex_match)
            except BaseException as err:
                exception = err
            else:
                parameters.append(parameter)
                continue
            
            await handle_command_exception(
                self,
                client,
                interaction_event,
                exception,
            )
            return
        
        command_coroutine = self._command_function(*parameters)
        
        try:
            await process_command_coroutine(client, interaction_event, False, command_coroutine)
        except BaseException as err:
            exception = err
        else:
            return
        
        await handle_command_exception(
            self,
            client,
            interaction_event,
            exception,
        )
        return
    
    def __hash__(self):
        """Returns the component command's hash value."""
        hash_value = 0
        
        string_custom_ids = self._string_custom_ids
        if (string_custom_ids is not None):
            hash_value ^= hash(string_custom_ids)
        
        regex_custom_ids = self._regex_custom_ids
        if (regex_custom_ids is not None):
            hash_value ^= hash(regex_custom_ids)
        
        command_function = self._command_function
        try:
            command_hash_value = hash(command_function)
        except KeyError:
            command_hash_value = object.__hash__(command_function)
        
        hash_value ^= command_hash_value
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether self equals to other."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self._command_function != other._command_function:
            return False
        
        if self._string_custom_ids != other._string_custom_ids:
            return False
        
        if self._regex_custom_ids != other._regex_custom_ids:
            return False
        
        if self._exception_handlers != other._exception_handlers:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the component command.
        
        Returns
        -------
        new : ``ComponentCommand``
        """
        new = object.__new__(type(self))
        new._command_function = self._command_function
        new._parameter_converters = self._parameter_converters
        new._string_custom_ids = self._string_custom_ids
        new._regex_custom_ids = self._regex_custom_ids
        new._parent_reference = None
        
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            exception_handlers = exception_handlers.copy()
        new._exception_handlers = exception_handlers
        
        return new


    def error(self, exception_handler=None, *, first=False):
        """
        Registers an exception handler to the ``SlasherApplicationCommandCategory``.
        
        Parameters
        ----------
        exception_handler : `None` or `CoroutineFunction`, Optional
            Exception handler to register.
        first : `bool`, Optional (Keyword Only)
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler / wrapper : `CoroutineFunction` / `functools.partial`
            If `exception_handler` is not given, returns a wrapper.
        """
        if exception_handler is None:
            return partial_func(_register_exception_handler, first)
        
        return self._register_exception_handler(exception_handler, first)
    
    
    def _register_exception_handler(self, exception_handler, first):
        """
        Registers an exception handler to the ``SlasherApplicationCommandCategory``.
        
        Parameters
        ----------
        exception_handler : `CoroutineFunction`
            Exception handler to register.
        first : `bool`
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler : `CoroutineFunction`
        """
        test_exception_handler(exception_handler)
        
        exception_handlers = self._exception_handlers
        if exception_handlers is None:
            self._exception_handlers = exception_handlers = []
        
        if first:
            exception_handlers.insert(0, exception_handler)
        else:
            exception_handlers.append(exception_handler)
        
        return exception_handler



__all__ = ('GuildWidget', 'GuildWidgetChannel', 'GuildWidgetUser',)

from ...backend.utils import cached_property

from ..bases import DiscordEntity
from ..user import Status

from ..http import urls as module_urls
from .guild import Guild

class GuildWidgetUser(DiscordEntity):
    """
    Represents an user object sent with a ``GuildWidget``'s data.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the guild widget user. Can be between `0` and `99`.
    activity_name : `None` or `str`
        The guild widget user's activity's name if applicable.
    avatar_url : `None` or `str`
        The guild widget user's avatar url if applicable.
    discriminator : `int`
        The guild widget user's discriminator.
    name : `str`
        The guild widget user's name.
    status : ``Status``
        The guild widget user's status.
    """
    __slots__ = ('activity_name', 'avatar_url', 'discriminator', 'name', 'status')
    
    def __init__(self, data):
        """
        Creates a new guild widget user from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Guild widget user data received with ``GuildWidget``'s.
        """
        self.name = data['username']
        self.id = int(data['id'])
        self.discriminator = int(data['discriminator'])
        self.avatar_url = data['avatar_url']
        self.status = Status.get(data['status'])
        try:
            activity_data = data['game']
        except KeyError:
            activity_name = None
        else:
            activity_name = activity_data['name']
        
        self.activity_name = activity_name
    
    
    @property
    def full_name(self):
        """
        The user's name with it's discriminator.
        
        Returns
        -------
        full_name : `str`
        """
        return f'{self.name}#{self.discriminator:0>4}'
    
    
    @property
    def mention(self):
        """
        The mention of the user.
        
        Returns
        -------
        mention : `str`
        """
        return f'<@{self.id}>'
    
    @property
    def mention_nick(self):
        """
        The mention to the user's nick.
        
        Returns
        -------
        mention : `str`
        
        Notes
        -----
        It actually has nothing to do with the user's nickname > <.
        """
        return f'<@!{self.id}>'
    
    
    def __repr__(self):
        """Returns the representation of the guild widget user."""
        return f'<{self.__class__.__name__} id={self.id}, name={self.full_name!r}>'


class GuildWidgetChannel(DiscordEntity):
    """
    Represents a ``GuildWidget``'s channel.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the guild widget channel.
    name : `str`
        The channel's name.
    position : `int`
        The channel's position.
    """
    __slots__ = ('name', 'position')
    
    def __init__(self, data):
        """
        Creates a new guild widget channel from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Guild widget channel data received with ``GuildWidget``'s.
        """
        self.id = int(data['id'])
        self.name = data['name']
        self.position = data['name']
    
    @property
    def mention(self):
        """
        The channel's mention.
        
        Returns
        -------
        mention : `str`
        """
        return f'<#{self.id}>'
    
    
    def __repr__(self):
        """Returns the guild widget channel's representation."""
        return f'<{self.__class__.__name__} id={self.id}, name={self.name!r}>'
    
    
    def __gt__(self, other):
        """
        Whether this guild widget channel has greater (visible) position than the other at their respective guild.
        """
        if type(self) is type(other):
            if self.position > other.position:
                return True
            
            if self.position == other.position:
                if self.id > other.id:
                    return True
            
            return False
        
        return NotImplemented
    
    def __ge__(self, other):
        """
        Whether this guild widget channel has greater or equal (visible) position than the other at their respective
        guild.
        """
        if type(self) is type(other):
            if self.position > other.position:
                return True
            
            if self.position == other.position:
                if self.id >= other.id:
                    return True
            
            return False
        
        return NotImplemented
    
    def __le__(self, other):
        """
        Whether this guild widget channel has lower or equal (visible) position than the other at their respective
        guild.
        """
        if type(self) is type(other):
            if self.position < other.position:
                return True
            
            if self.position == other.position:
                if self.id <= other.id:
                    return True
            
            return False
        
        return NotImplemented
    
    def __lt__(self, other):
        """
        Whether this guild widget channel has lower (visible) position than the other at their respective guild.
        """
        if type(self) is type(other):
            if self.position < other.position:
                return True
            
            if self.position == other.position:
                if self.id < other.id:
                    return True
            
            return False
        
        return NotImplemented

class GuildWidget(DiscordEntity):
    """
    Represents a ``Guild``'s widget.
    
    Attributes
    ----------
    _cache : `dict` of (`str`, `Any`) items
        Internal cache used by cached properties.
    _data : `dict` of (`str`, `Any`) items
        The data sent by Discord and used by the cached properties of the guild widget instances.
    guild : ``Guild``
        The owner guild of the widget.
    """
    __slots__ = ('_cache', '_data', 'guild',)
    
    def __init__(self, data):
        """
        Creates a new guild widget.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The requested guild widget data.
        """
        self.guild = Guild.precreate(int(data['id']), name=data['name'])
        self._data = data
        self._cache = {}
    
    json_url = property(module_urls.guild_widget_json_url)
    
    @property
    def id(self):
        """
        The unique identifier number of the guild widget's guild.
        
        Returns
        -------
        id : `int`
        """
        return self.guild.id
    
    @property
    def name(self):
        """
        The name of the guild widget's guild.
        
        Returns
        -------
        name : `str`
        """
        return self.guild.name
    
    @property
    def invite_url(self):
        """
        The guild widget's invite url if applicable.
        
        Returns
        -------
        invite_url : `None` or `str`
        """
        return self._data.get('instant_invite', None)
    
    @property
    def approximate_online_count(self):
        """
        Estimated online count of the respective guild.
        
        Returns
        -------
        approximate_online_count : `int`
        """
        return self._data['presence_count']
    
    @cached_property
    def users(self):
        """
        Online users received with the guild widget.
        
        Returns
        -------
        users : `list` of ``GuildWidgetUser``
        """
        return [GuildWidgetUser(GWU_data) for GWU_data in self._data['members']]

    @cached_property
    def channels(self):
        """
        Voice channels received with the guild widget.
        
        Returns
        -------
        users : `list` of ``GuildWidgetChannel``
        """
        return [GuildWidgetChannel(GWC_data) for GWC_data in self._data['channels']]
    
    def __repr__(self):
        """Returns the representation of the guild widget."""
        return f'<{self.__class__.__name__} of guild {self.guild.name}>'



__all__ = ('ChannelPrivate', 'ChannelGroup', )

from ...backend.utils import copy_docs
from ...backend.export import export

from ..bases import IconSlot, ICON_TYPE_NONE
from ..core import CHANNELS
from ..permission.permission import PERMISSION_NONE, PERMISSION_PRIVATE, PERMISSION_PRIVATE_BOT, \
    PERMISSION_GROUP, PERMISSION_GROUP_OWNER
from ..user import User, ZEROUSER, create_partial_user_from_id
from ..preconverters import preconvert_snowflake, preconvert_str
from ..http import urls as module_urls

from .channel_base import ChannelBase
from .channel_text_base import ChannelTextBase

@export
class ChannelPrivate(ChannelBase, ChannelTextBase):
    """
    Represents a private (/ direct message) channel.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the channel.
    _message_history_collector :  `None` or ``MessageHistoryCollector``
        Collector for the channel's message history.
    _message_keep_limit : `int`
        The channel's own limit of how much messages it should keep before removing their reference.
    message_history_reached_end : `bool`
        Whether the channel's message's are loaded till their end. If the channel's message history reached it's end
        no requests will be requested to get older messages.
    messages : `deque` of ``Message`` objects
        The channel's message history.
    users : `list` of ``ClientUserBase`` objects
        The channel's recipient.
    
    Class Attributes
    ----------------
    DEFAULT_TYPE : `int` = `1`
        The default type, what the channel represents.
    INTERCHANGE : `tuple` of `int` = `(1,)`
        Defines to which channel type this channel's type can be interchanged. The channel's direct type must be of
        them.
    type : `int` = `1`
        The channel's Discord side type.
    """
    __slots__ = ('users',) # private related
    
    DEFAULT_TYPE = 1
    INTERCHANGE = (1,)
    type = 1
    
    def __new__(cls, data, client, guild_id):
        """
        Creates a private channel from the channel data received from Discord. If the channel already exists and if it
        is partial, then updates it.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data receive from Discord.
        client : `None` or ``Client``
            The client, who received the channel's data, if any.
        guild_id : `int`
            The channel's guild's identifier.
        """
        channel_id = int(data['id'])
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = channel_id
            CHANNELS[channel_id] = self
            self._messageable_init()
            self.users = users = []
        else:
            users = self.users
            if len(users) == 2:
                return self
            
            users.clear()
        
        for user_data in data['recipients']:
            user = User(user_data)
            users.append(user)
        
        if (client is not None):
            if client not in users:
                users.append(client)
            
            users.sort()
            
            user = users[0]
            if user is client:
                user = users[1]
                
            client.private_channels[user.id] = self
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelBase._create_empty)
    def _create_empty(cls, channel_id, channel_type, guild_id):
        self = super(ChannelPrivate, cls)._create_empty(channel_id, channel_type, guild_id)
        self._messageable_init()
        self.users = []
        
        return self
    
    
    @classmethod
    def _create_dataless(cls, channel_id):
        """
        Creates a private channel from a channel id. Might be called by events, when a message's channel is not found
        and it is a private channel.
        
        Parameters
        ----------
        channel_id : `int`
            The channel's respective id.
        
        Returns
        -------
        channel : ``ChannelPrivate``
            The created channel.
        """
        self = object.__new__(cls)
        self._messageable_init()
        self.id = channel_id
        self.users = []
        CHANNELS[channel_id] = self
        
        return self
    
    def _finish_dataless(self, client, user):
        """
        Finishes the initialization of the channel after a ``.create_dataless`` call.
        
        Parameters
        ----------
        client : ``Client``
            The client recipient of the channel.
        user : ``ClientUserBase``
            The other recipient of the channel.
        """
        users = self.users
        users.append(client)
        users.append(user)
        users.sort()
        
        client.private_channels[user.id] = self
    
    @copy_docs(ChannelBase._get_processed_name)
    def _get_processed_name(self):
        users = self.users
        if users:
            name = f'Direct Message {users[0].full_name} with {users[1].full_name}'
        else:
            name = f'Direct Message (partial)'
        return name
    
    
    def _delete(self, client):
        """
        Removes the channel's references.
        
        Used when the channel is deleted.
        
        Parameters
        ----------
        client : ``Client``
            The client, who's private channel was deleted.
        """
        users = self.users
        if client is users[0]:
            user = users[1]
        else:
            user = users[0]
        
        del client.private_channels[user.id]
    
    
    name = property(_get_processed_name)
    copy_docs(ChannelBase.name)(name)
    
    display_name = property(_get_processed_name)
    copy_docs(ChannelBase.display_name)(display_name)
    
    @copy_docs(ChannelBase.permissions_for)
    def permissions_for(self, user):
        if user in self.users:
            if user.is_bot:
                return PERMISSION_PRIVATE_BOT
            else:
                return PERMISSION_PRIVATE
            
        return PERMISSION_NONE
    
    
    cached_permissions_for = permissions_for
    
    
    @classmethod
    def precreate(cls, channel_id, **kwargs):
        """
        Precreates the channel by creating a partial one with the given parameters. When the channel is loaded
        the precrated channel will be picked up. If an already existing channel would be precreated, returns that
        instead and updates that only, if that is a partial channel.
        
        Parameters
        ----------
        channel_id : `int` or `str`
            The channel's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the channel.
        
        Returns
        -------
        channel : ``ChannelPrivate``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        channel_id = preconvert_snowflake(channel_id, 'channel_id')
        
        if kwargs:
            raise ValueError(f'Unused or unsettable attributes: {kwargs}')
        
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = cls._create_empty(channel_id, cls.type, 0)
            CHANNELS[channel_id] = self
        
        return self



@export
class ChannelGroup(ChannelBase, ChannelTextBase):
    """
    Represents a group channel.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the channel.
    _message_history_collector :  `None` or ``MessageHistoryCollector``
        Collector for the channel's message history.
    _message_keep_limit : `int`
        The channel's own limit of how much messages it should keep before removing their reference.
    message_history_reached_end : `bool`
        Whether the channel's message's are loaded till their end. If the channel's message history reached it's end
        no requests will be requested to get older messages.
    messages : `deque` of ``Message`` objects
        The channel's message history.
    users : `list` of ``ClientUserBase`` objects
        The channel's recipient.
    icon_hash : `int`
        The channel's icon's hash in `uint128`.
    icon_type : ``iconType``
        The channel's icon's type.
    name : `str`
        The channel's display name. Can be empty string if the channel has no name.
    owner_id : `int`
        The group channel's owner's id.
    
    Class Attributes
    ----------------
    DEFAULT_TYPE : `int` = `3`
        The default type, what the channel represents.
    INTERCHANGE : `tuple` of `int` = `(3,)`
        Defines to which channel type this channel's type can be interchanged. The channel's direct type must be of
        them.
    type : `int` = `3`
        The channel's Discord side type.
    """
    __slots__ = ('users', # private channel related
        'name', 'owner_id',) # group channel related
    
    icon = IconSlot('icon', 'icon', module_urls.channel_group_icon_url, module_urls.channel_group_icon_url_as)
    
    DEFAULT_TYPE = 3
    INTERCHANGE = (3,)
    type = 3
    
    def __new__(cls, data, client, guild_id):
        """
        Creates a channel from the channel data received from Discord. If the channel already exists and if it is
        partial, then updates it.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data receive from Discord.
        client : `None` or ``Client``
            The client, who received the channel's data, if any.
        guild_id : `int`
            The channel's guild's identifier.
        """
        channel_id = int(data['id'])
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = channel_id
            CHANNELS[channel_id] = self
            self._messageable_init()
            self.users = users = []
        else:
            users = self.users
            if len(users) == len(data['recipients']):
                return self
            
            users.clear()
        
        name = data.get('name', None)
        self.name = '' if name is None else name
        self._set_icon(data)
        self.owner_id = int(data['owner_id'])
        
        for user_data in data['recipients']:
            user = User(user_data)
            users.append(user)
        
        users.sort()
        
        if (client is not None):
            client.group_channels[channel_id] = self
        
        return self
    
    
    @property
    def owner(self):
        """
        Returns the group channel's owner.
        
        Returns
        -------
        owner : ``ClientUserBase``
            Defaults to `ZEROUSER`.
        """
        owner_id = self.owner_id
        if owner_id:
            owner = create_partial_user_from_id(owner_id)
        else:
            owner = ZEROUSER
        return owner
    
    
    @classmethod
    @copy_docs(ChannelBase._from_partial_data)
    def _from_partial_data(cls, data, channel_id, guild_id):
        self = super(ChannelGroup, cls)._from_partial_data(data, channel_id, guild_id)
        
        name = data.get('name', None)
        if (name is not None):
            self.name = name
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelBase._create_empty)
    def _create_empty(cls, channel_id, channel_type, guild_id):
        self = super(ChannelGroup, cls)._create_empty(channel_id, channel_type, guild_id)
        self._messageable_init()
        
        self.users = []
        
        self.name = None
        self.owner_id = 0
        
        self.icon_hash = 0
        self.icon_type = ICON_TYPE_NONE
        
        return self
    
    
    def _delete(self, client):
        """
        Removes the channel's references.
        
        Used when the channel is deleted.
        
        Parameters
        ----------
        client : ``Client``
            The client, who's private channel was deleted.
        """
        del client.group_channels[self.id]
    
    
    @copy_docs(ChannelBase._update_attributes)
    def _update_attributes(self, data):
        name = data.get('name', None)
        self.name = '' if name is None else name
        
        self._set_icon(data)
        
        self.owner_id = int(data['owner_id'])
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the channel and returns it's overwritten old attributes as a `dict` with a `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
            
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        +---------------+---------------------------------------+
        | Keys          | Values                                |
        +===============+=======================================+
        | icon          | ``Icon``                              |
        +---------------+---------------------------------------+
        | name          | `str`                                 |
        +---------------+---------------------------------------+
        | owner_id      | `int`                                 |
        +---------------+---------------------------------------+
        """
        old_attributes = {}
        
        name = data.get('name', None)
        if name is None:
            name = ''
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        self._update_icon(data, old_attributes)
        
        owner_id = int(data['owner_id'])
        if self.owner_id != owner_id:
            old_attributes['owner_id'] = self.owner_id
            self.owner_id = owner_id
        
        return old_attributes
    
    @copy_docs(ChannelBase._get_processed_name)
    def _get_processed_name(self):
        name = self.name
        if name:
            return name
        
        users = self.users
        if users:
            return ', '.join([user.name for user in users])
        
        return 'Unnamed'
    
    display_name = property(_get_processed_name)
    copy_docs(ChannelBase.display_name)(display_name)
    
    @copy_docs(ChannelBase.permissions_for)
    def permissions_for(self, user):
        if self.owner_id == user.id:
            return PERMISSION_GROUP_OWNER
        elif user in self.users:
            return PERMISSION_GROUP
        else:
            return PERMISSION_NONE
    
    cached_permissions_for = permissions_for
    
    
    @classmethod
    def precreate(cls, channel_id, **kwargs):
        """
        Precreates the channel by creating a partial one with the given parameters. When the channel is loaded
        the precrated channel will be picked up. If an already existing channel would be precreated, returns that
        instead and updates that only, if that is a partial channel.
        
        Parameters
        ----------
        channel_id : `int` or `str`
            The channel's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the channel.
        
        Other Parameters
        ----------------
        name : `str`, Optional (Keyword only)
            The channel's ``.name``.
        icon : `int`, Optional (Keyword only)
            The channel's ``.icon``.
        owner_id : `int`, Optional (Keyword only)
            The channel's owner's id.
        
        Returns
        -------
        channel : ``ChannelGroup``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        channel_id = preconvert_snowflake(channel_id, 'channel_id')
        
        if kwargs:
            processable = []
            
            try:
                name = kwargs.pop('name')
            except KeyError:
                pass
            else:
                name = preconvert_str(name, 'name', 2, 100)
                processable.append(('name', name))
            
            cls.icon.preconvert(kwargs, processable)
            
            try:
                owner_id = kwargs.pop('owner_id')
            except KeyError:
                pass
            else:
                owner_id = preconvert_snowflake(owner_id, 'owner_id')
                processable.append(('owner_id', owner_id))
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}')
        
        else:
            processable = None
        
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = cls._create_empty(channel_id, cls.type, 0)
            CHANNELS[channel_id] = self
            
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self



from .activity_change import *
from .client_user_base import *
from .flags import *
from .guild_profile import *
from .helpers import *
from .preinstanced import *
from .thread_profile import *
from .user import *
from .user_base import *
from .utils import *
from .voice_state import *


__all__ = (
    *activity_change.__all__,
    *client_user_base.__all__,
    *flags.__all__,
    *guild_profile.__all__,
    *helpers.__all__,
    *preinstanced.__all__,
    *thread_profile.__all__,
    *user.__all__,
    *user_base.__all__,
    *utils.__all__,
    *voice_state.__all__,
)



__all__ = ('Button', 'ButtonStyle', 'Row', 'Select', 'Option')

from ...discord.exceptions import DiscordException, ERROR_CODES
from ...discord.core import APPLICATION_ID_TO_CLIENT

from ...discord.interaction import ComponentButton as Button, ComponentRow as Row, ComponentSelect as Select, \
    ComponentSelectOption as Option, ButtonStyle


async def acknowledge_component_interaction(interaction_event):
    """
    Acknowledges the given `interaction_event`. If error occurs, calls `client.events.error`.
    
    This function is a coroutine.
    
    Parameter
    ---------
    interaction_event : ``InteractionEvent``
        The component interaction event to acknowledge.
    """
    try:
        client = APPLICATION_ID_TO_CLIENT[interaction_event.application_id]
    except KeyError:
        return
        
    try:
        await client.interaction_component_acknowledge(interaction_event)
    except BaseException as err:
        if isinstance(err, ConnectionError):
            # No Internet connection
            return
        
        if isinstance(err, DiscordException) and (err.code == ERROR_CODES.unknown_interaction):
            # We timed out, bad connection.
            return
        
        await client.events.error(client, f'acknowledge_component_interaction', err)



__all__ = ('CHANNEL_MENTION_RP', 'DATETIME_FORMAT_CODE', 'DISCORD_EPOCH', 'EMOJI_NAME_RP', 'EMOJI_RP', 'Gift', 'ID_RP',
    'IS_MENTION_RP', 'REACTION_RP', 'ROLE_MENTION_RP', 'Relationship', 'USER_MENTION_RP', 'Unknown', 'cchunkify',
    'chunkify', 'datetime_to_id', 'datetime_to_timestamp', 'datetime_to_unix_time', 'elapsed_time', 'escape_markdown',
    'filter_content', 'id_to_datetime', 'id_to_time', 'id_to_unix_time', 'is_id', 'is_invite_code', 'is_mention',
    'is_role_mention', 'is_url', 'is_user_mention', 'mention_channel_by_id', 'mention_role_by_id',
    'mention_user_by_id', 'mention_user_nick_by_id', 'now_as_id', 'parse_message_reference', 'parse_rdelta',
    'parse_tdelta', 'random_id', 'sanitize_content', 'sanitize_mentions', 'time_to_id', 'unix_time_to_id')

import sys, warnings
from random import random
from re import compile as re_compile, I as re_ignore_case, U as re_unicode
from datetime import datetime, timedelta, timezone
from base64 import b64encode
from time import time as time_now
from math import floor
from email._parseaddr import _parsedate_tz as parse_date_timezone
from functools import partial as partial_func

try:
    from dateutil.relativedelta import relativedelta
except ImportError:
    relativedelta = None

from ..backend.export import export, include
from ..backend.utils import modulize, IS_UNIX, set_docs

from .bases import DiscordEntity
from .core import USERS, CHANNELS, ROLES

MESSAGE_JUMP_URL_RP = include('MESSAGE_JUMP_URL_RP')
create_partial_user_from_id = include('create_partial_user_from_id')
RelationshipType = include('RelationshipType')

DATETIME_FORMAT_CODE = '%Y-%m-%d %H:%M:%S'

def endswith_xFFxD9(data):
    """
    Checks whether the given data endswith `b'\xD9\xFF'` ignoring empty bytes at the end of it.
    
    Parameters
    ----------
    data : `bytes-like`
    
    Returns
    -------
    result : `bool`
    """
    index = len(data)-1
    while index > 1:
        actual = data[index]
        if actual == b'\xD9'[0] and data[index-1] == b'\xFF'[0]:
            return True
        
        if actual:
            return False
        
        index -= 1
        continue


def get_image_media_type(data):
    """
    Gets the given raw image data's extension and returns it.
    
    Parameters
    ----------
    data : `bytes-like`
        Image data.
    
    Returns
    -------
    media_type : `str`
    """
    if data.startswith(b'\x89\x50\x4E\x47\x0D\x0A\x1A\x0A'):
        media_type = 'image/png'
    elif data.startswith(b'\xFF\xD8') and endswith_xFFxD9(data):
        media_type = 'image/jpeg'
    elif data.startswith(b'\x47\x49\x46\x38\x37\x61') or data.startswith(b'\x47\x49\x46\x38\x39\x61'):
        media_type = 'image/gif'
    elif data.startswith(b'{') and data.endswith(b'}'):
        media_type = 'application/json'
    else:
        media_type = ''
    
    return media_type


MEDIA_TYPE_TO_EXTENSION = {
    'image/png': 'png',
    'image/jpeg': 'jpg',
    'image/gif': 'gif',
    'application/json': 'json',
}


def image_to_base64(data):
    """
    Converts a bytes image to a base64 one.
    
    Parameter
    ----------
    data : `bytes-like`
        Image data.
    
    Returns
    -------
    base64 : `str`
    
    Raises
    ------
    ValueError
        If `ext` was not given and the given `data`'s image format is not any of the expected ones.
    """
    if data.startswith(b'\x89\x50\x4E\x47\x0D\x0A\x1A\x0A'):
        media_type = 'image/png'
    elif data.startswith(b'\xFF\xD8') and endswith_xFFxD9(data):
        media_type = 'image/jpeg'
    elif data.startswith(b'\x47\x49\x46\x38\x37\x61') or data.startswith(b'\x47\x49\x46\x38\x39\x61'):
        media_type = 'image/gif'
    else:
        raise ValueError('Unsupported image type given.')
    
    return ''.join(['data:', media_type, ';base64,', b64encode(data).decode('ascii')])


DISCORD_EPOCH = 1420070400000

# example dates:
# "2016-03-31T19:15:39.954000+00:00"
# "2019-04-28T15:14:38+00:00"
# at edit:
# "2019-07-17T18:52:50.758993+00:00" #this is before desuppress!
# at desuppress:
# "2019-07-17T18:52:50.758000+00:00"

PARSE_TIMESTAMP_RP = re_compile('(\\d{4})-(\\d{2})-(\\d{2})T(\\d{2}):(\\d{2}):(\\d{2})(?:\\.(\\d{3})?)?.*')

def timestamp_to_datetime(timestamp):
    """
    Parses the given timestamp.
    
    The expected timestamp formats are the following:
        - `2019-04-28T15:14:38+00:00`
        - `2019-07-17T18:52:50.758993+00:00`
        - `2019-07-17T18:52:50.758000+00:00`
        
    Discord might give timestamps with different accuracy, so we use an optimal middle way at parsing them.
    Some events depend on timestamp accuracy, so we really do not want to be wrong about them, or it might cause
    same internal derpage.
    
    If parsing a timestamp failed, the start of the discord epoch is returned and an error message is given at
    `sys.stderr`.
    
    Parameters
    ----------
    timestamp : `str`
        The timestamp to parse.
    
    Returns
    -------
    time : `datetime`
    
    Notes
    -----
    I already noted that timestamp formats are inconsistent, but even our baka Chiruno could have fix it...
    """
    parsed = PARSE_TIMESTAMP_RP.fullmatch(timestamp)
    if parsed is None:
        sys.stderr.write(f'Cannot parse timestamp: `{timestamp}`, returning `DISCORD_EPOCH_START`\n')
        return DISCORD_EPOCH_START
    
    year = int(parsed.group(1))
    month = int(parsed.group(2))
    day = int(parsed.group(3))
    hour = int(parsed.group(4))
    minute = int(parsed.group(5))
    second = int(parsed.group(6))
    micro = parsed.group(7)
    
    if micro is None:
        micro = 0
    else:
        micro = int(micro)
    
    return datetime(year, month, day, hour, minute, second, micro)


def datetime_to_timestamp(date_time):
    """
    Converts the given datetime to it's timestamp representation.
    
    Parameters
    ----------
    date_time : `datetime`
        The datetime to convert to timestamp.
    
    Returns
    -------
    timestamp : `str`
    """
    return date_time.isoformat()


@export
def id_to_time(id_):
    """
    `id_to_time`\'s is deprecated, and will be removed in 2021 November. Please use ``id_to_datetime`` instead.
    """
    warnings.warn(
        f'`id_to_time`\'s is deprecated, and will be removed in 2021 November. '
        f'Please use `id_to_datetime` instead.',
        FutureWarning)
    return id_to_datetime(id_)


@export
def id_to_datetime(id_):
    """
    Converts the given id to datetime.
    
    Parameters
    ----------
    id_ : `int`
        Unique identifier number of a Discord entity.
    
    Returns
    -------
    date_time : `datetime`
    """
    return datetime.utcfromtimestamp(((id_>>22)+DISCORD_EPOCH)/1000.)


DISCORD_EPOCH_START = id_to_datetime(0)

def id_to_unix_time(id_):
    """
    Converts the given id to unix time.
    
    Parameters
    ----------
    id_ : `int`
        Unique identifier number of a Discord entity.
    
    Returns
    -------
    unix_time : `int`
    """
    return ((id_>>22)+DISCORD_EPOCH)//1000


def unix_time_to_id(unix_time):
    """
    Converts the given unix time to id.
    
    Parameters
    ----------
    unix_time : `int`, `float`
        The unix time to convert to id.
    
    Returns
    -------
    id_ : `int`
    """
    return (floor(unix_time*1000.)-DISCORD_EPOCH)<<22


def unix_time_to_datetime(unix_time):
    """
    Converts the given unix time to datetime.
    
    Parameters
    ----------
    unix_time : `int`, `float`
        The unix time to convert to datetime.
    
    Returns
    -------
    date_time : `datetime`
    """
    try:
        return datetime.utcfromtimestamp(unix_time/1000.0)
    except ValueError:
        # Can happen if max or min year is passed.
        if unix_time >= UNIX_TIME_MAX:
            return DATETIME_MAX
        else:
            return DATETIME_MIN


def time_to_id(time):
    """
    `id_to_time`\'s is deprecated, and will be removed in 2021 November. Please use ``id_to_datetime`` instead.
    """
    warnings.warn(
        f'`id_to_time`\'s is deprecated, and will be removed in 2021 November. '
        f'Please use `id_to_datetime` instead.',
        FutureWarning)
    return datetime_to_id(time)


def datetime_to_id(date_time):
    """
    Converts the given time to it's respective discord identifier number.
    
    Parameters
    ----------
    date_time : `datetime`
        The datetime to convert to Discord identifier.
    
    Returns
    -------
    id_ `int`
    """
    return (floor(date_time.timestamp()*1000.)-DISCORD_EPOCH)<<22


if IS_UNIX:
    def datetime_to_unix_time(date_time):
        return floor(date_time.timestamp()*1000.0)
else:
    def datetime_to_unix_time(date_time):
        try:
            return floor(date_time.timestamp()*1000.0)
        except OSError:
            if date_time <= DATETIME_MIN:
                return UNIX_TIME_MIN
            else:
                return UNIX_TIME_MAX


set_docs(datetime_to_unix_time,
    """
    Converts the given time to it's unix time value.
    
    Parameters
    ----------
    date_time : `datetime`
        The datetime to convert to unix time.
    
    Returns
    -------
    unit_time : `int`
    """
)


UNIX_TIME_MIN = 0
DATETIME_MIN = unix_time_to_datetime(UNIX_TIME_MIN)

DATETIME_MAX = datetime(year=3000, month=1, day=1)
UNIX_TIME_MAX = datetime_to_unix_time(DATETIME_MAX)


def random_id():
    """
    Generates a random Discord identifier number what's datetime value did not surpass the current time.
    
    Returns
    -------
    id_ `int`
    """
    return ((floor(time_now()*1000.)-DISCORD_EPOCH)<<22)+floor(random()*4194304.0)


def log_time_converter(value):
    """
    Converts the given value to it's snowflake representation.
    
    Parameters
    ----------
    value : `int`, ``DiscordEntity`` instance or `datetime`
        If the value is given as `int`, returns it. If given as a ``DiscordEntity``, then returns it's id and if it
        is given as a `datetime` object, then converts that to snowflake then returns it.
    
    Returns
    -------
    snowflake : `int`
    
    Raises
    ------
    TypeError
        If the given `value`'s type is not any of the expected ones.
    """
    if isinstance(value, int):
        return value
    
    if isinstance(value, DiscordEntity):
        return value.id
    
    if type(value) is datetime:
        return time_to_id(value)
    
    raise TypeError(f'Expected `int`, `{DiscordEntity.__name__}` instance, or a `datetime` object, got '
        f'`{value.__class__.__name__}`.')

APPLICATION_COMMAND_NAME_RP = re_compile('[a-zA-Z0-9_\-]{1,32}')

ID_RP = re_compile('(\d{7,21})')
IS_MENTION_RP = re_compile('@(?:everyone|here)|<(?:@[!&]?|#|/[a-zA-Z0-9_\-]{3,32}:)\d{7,21}>')

USER_MENTION_RP = re_compile('<@!?(\d{7,21})>')
CHANNEL_MENTION_RP = re_compile('<#(\d{7,21})>')
ROLE_MENTION_RP = re_compile('<@&(\d{7,21})>')
APPLICATION_COMMAND_MENTION_RP = re_compile('</([a-zA-Z0-9_\-]{3,32}):(\d{7,21})>')

EMOJI_RP = re_compile('<(a)?:([a-zA-Z0-9_]{2,32})(?:~[1-9])?:(\d{7,21})>')
REACTION_RP = re_compile('([a-zA-Z0-9_]{2,32}):(\d{7,21})')
EMOJI_NAME_RP = re_compile(':?([a-zA-Z0-9_\\-~]{1,32}):?')
FILTER_RP = re_compile('("(.+?)"|\S+)')
INVITE_CODE_RP = re_compile('([a-zA-Z0-9-]+)')


def mention_channel_by_id(channel_id):
    """
    Mentions the channel by it's identifier.
    
    Parameters
    ----------
    channel_id : `int`
        The channel's identifier.
    
    Returns
    -------
    channel_mention : `str`
    """
    return f'<#{channel_id}>'


def mention_role_by_id(role_id):
    """
    Mentions the role by it's identifier.
    
    Parameters
    ----------
    role_id : `int`
        The role's identifier.
    
    Returns
    -------
    role_mention : `str`
    """
    return f'<@&{role_id}>'


def mention_user_by_id(user_id):
    """
    Mentions the user by it's identifier.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    Returns
    -------
    user_mention : `str`
    """
    return f'<@{user_id}>'


def mention_user_nick_by_id(user_id):
    """
    Mentions the user's "nick" by the user's identifier.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    Returns
    -------
    user_nick_mention : `str`
    """
    return f'<@!{user_id}>'


def is_valid_application_command_name(name):
    """
    Returns whether the given application command name is valid.
    
    Parameters
    ----------
    name : `str`
        The name of the application command.
    
    Returns
    -------
    valid : `bool`
    """
    return (APPLICATION_COMMAND_NAME_RP.fullmatch(name) is not None)


def is_id(text):
    """
    Returns whether the given text is a valid snowflake.
    
    Parameters
    ----------
    text : `str`
        The text to check.
    
    Returns
    -------
    result : `bool`
    """
    return ID_RP.fullmatch(text) is not None


def is_mention(text):
    """
    Returns whether the given text is a mention.
    
    Parameters
    ----------
    text : `str`
        The text to check.
    
    Returns
    -------
    result : `bool`
    """
    return IS_MENTION_RP.fullmatch(text) is not None


def is_user_mention(text):
    """
    Returns whether the given text mentions a user.
    
    Parameters
    ----------
    text : `str`
        The text to check.
    
    Returns
    -------
    result : `str`
    """
    return USER_MENTION_RP.fullmatch(text) is not None


def is_channel_mention(text):
    """
    Returns whether the given text mentions a channel.
    
    Parameters
    ----------
    text : `str`
        The text to check.
    
    Returns
    -------
    result : `bool`
    """
    return CHANNEL_MENTION_RP.fullmatch(text) is not None


def is_role_mention(text):
    """
    Returns whether the given text mentions a role.
    
    Parameters
    ----------
    text : `str`
        The text to check.
    
    Returns
    -------
    result : `bool`
    """
    return ROLE_MENTION_RP.fullmatch(text) is not None


def is_application_command_mention(text):
    """
    Returns whether the given text is an interaction command mention.
    
    Parameters
    ----------
    text : `str`
        The text to check.
    
    Returns
    -------
    result : `bool`
    """
    return (APPLICATION_COMMAND_MENTION_RP.fullmatch(text) is not None)


def is_invite_code(text):
    """
    Returns whether the given text is an invite code
    
    Parameters
    ----------
    text : `str`
        The text to check.
    
    Returns
    -------
    result : `bool`
    """
    return (INVITE_CODE_RP.fullmatch(text) is not None)


def now_as_id():
    """
    Returns the current time as a Discord snowflake.
    
    Returns
    -------
    snowflake : `int`
    """
    return (floor(time_now()*1000.)-DISCORD_EPOCH)<<22


def filter_content(content):
    """
    Filters the given content to parts separated with spaces. Parts surrounded with `"` character will count as one
    even if they contain spaces.
    
    Parameters
    ----------
    content : `str`
        The text to check.
    
    Returns
    -------
    parts : `list` of `str`
    """
    return [match[1] or match[0] for match in FILTER_RP.findall(content)]


def chunkify(lines, limit=2000):
    """
    Creates chunks of strings from the given lines.
    
    Parameters
    ----------
    lines : `list` of `str`
        Lines of text to be chunkified.
    limit : `int`, Optional
        The maximal length of a generated chunk.
    
    Returns
    -------
    result : `list` of `str`
    
    Raises
    ------
    ValueError`
        If limit is less than `500`.
    """
    if limit < 500:
        raise ValueError(f'Minimal limit should be at least 500, got {limit!r}.')
    
    result = []
    chunk_length = 0
    chunk = []
    for line in lines:
        while True:
            ln = len(line)+1
            if chunk_length+ln > limit:
                position = limit-chunk_length
                if position < 250:
                    result.append('\n'.join(chunk))
                    chunk.clear()
                    chunk.append(line)
                    chunk_length=ln
                    break
                
                position = line.rfind(' ', position-250, position-3)
                if position == -1:
                    position = limit-chunk_length-3
                    post_part = line[position:]
                else:
                    post_part = line[position+1:]
                
                pre_part = line[:position]+'...'
                
                chunk.append(pre_part)
                result.append('\n'.join(chunk))
                chunk.clear()
                if len(post_part) > limit:
                    line = post_part
                    chunk_length = 0
                    continue
                
                chunk.append(post_part)
                chunk_length = len(post_part)+1
                break
            
            chunk.append(line)
            chunk_length += ln
            break
    
    result.append('\n'.join(chunk))
    
    return result


def cchunkify(lines, lang='', limit=2000):
    """
    Creates code block chunks from the given lines.
    
    Parameters
    ----------
    lines : `list` of `str`
        Lines of text to be chunkified.
    lang : `str`, Optional
        Language prefix of the code-block.
    limit : `int`, Optional
        The maximal length of a generated chunk.
    
    Returns
    -------
    result : `list` of `str`
    
    Raises
    ------
    ValueError`
        If limit is less than `500`.
    """
    if limit < 500:
        raise ValueError(f'Minimal limit should be at least 500, got {limit!r}.')
    
    starter = f'```{lang}'
    limit = limit-len(starter)-5
    
    result = []
    chunk_length = 0
    chunk = [starter]
    for line in lines:
        while True:
            ln = len(line)+1
            if chunk_length+ln > limit:
                position = limit-chunk_length
                if position < 250:
                    chunk.append('```')
                    result.append('\n'.join(chunk))
                    chunk.clear()
                    chunk.append(starter)
                    chunk.append(line)
                    chunk_length = ln
                    break
                
                position = line.rfind(' ', position-250, position-3)
                if position == -1:
                    position = limit-chunk_length-3
                    post_part = line[position:]
                else:
                    post_part = line[position+1:]
                
                pre_part = line[:position]+'...'
                
                chunk.append(pre_part)
                chunk.append('```')
                result.append('\n'.join(chunk))
                chunk.clear()
                chunk.append(starter)
                
                if len(post_part) > limit:
                    line = post_part
                    chunk_length = 0
                    continue
                
                chunk.append(post_part)
                chunk_length = len(post_part)+1
                break
            
            chunk.append(line)
            chunk_length += ln
            break
    
    if len(chunk)>1:
        chunk.append('```')
        result.append('\n'.join(chunk))
    
    return result

if relativedelta is None:
    elapsed_time = None
else:
    def elapsed_time(delta, limit=3, names=(
            ('year', 'years',),
            ('month', 'months'),
            ('day', 'days', ),
            ('hour', 'hours'),
            ('minute', 'minutes'),
            ('second', 'seconds'),)):
        """
        Generates an elapsed time formula from the given time delta.
        
        Parameters
        ----------
        delta : `datetime` or `relativedelta`
            The time delta. If given as `datetime`, then the delta will be based on the difference between the given
            datetime and the actual time. If given as `relativedelta`, then that will be used directly.
        limit : `int`, Optional
            The maximal amount of connected time units. Defaults to `3`.
        names : `iterable` of `tuple` (`str`, `str`), Optional
            The names of the time units starting from years. Each element of the iterable should yield a `tuple` of two
            `str` elements. The first should be always the singular form of the time unit's name and the second the
            plural. Defaults to the time units' names in engrisssh.
        
        Returns
        -------
        result : `str`
        
        Raises
        ------
        TypeError
            If delta was neither passed as `datetime` or as `relativedelta` instance.
        """
        if isinstance(delta, datetime):
            delta = relativedelta(datetime.utcnow(), delta)
        elif isinstance(delta, relativedelta):
            pass
        else:
            raise TypeError(f'Expected, `relativedelta` or `datetime`, got {delta.__class__.__name__}.')
        
        parts = []
        for value, name_pair in zip((delta.years, delta.months, delta.days, delta.hours, delta.minutes, delta.seconds), names):
            if limit == 0:
                break
            
            if value < 0:
                 value = -value
            elif value == 0:
                continue
            
            parts.append(str(value))
            parts.append(' ')
            
            name = name_pair[value!=1]
            
            parts.append(name)
            parts.append(', ')
            
            limit -= 1
        
        if parts:
            del parts[-1]
            result = ''.join(parts)
        else:
            result = f'0 {names[5][0]}'
        
        return result


class Relationship:
    """
    Represents a Discord relationship.
    
    Attributes
    ----------
    type : ``RelationshipType``
        The relationship's type.
    user : ``ClientUserBase``
        The target user of the relationship.
    """
    __slots__ = ('type', 'user',)
    def __init__(self, client, data, user_id):
        """
        Creates a relationship instance with the given parameters.
        
        Parameters
        ----------
        client : ``Client``
            The client, who's relationship is created.
        data : `dict` of (`str`, `Any`)
            Relationship data.
        user_id : `int`
            The relationship's target user's id.
        """
        self.user = create_partial_user_from_id(user_id)
        self.type = RelationshipType.get(data['type'])
        client.relationships[user_id] = self
    
    def __repr__(self):
        """Returns the representation of the relationship."""
        return f'<{self.__class__.__name__} {self.type.name} user={self.user.full_name!r}>'


class Unknown(DiscordEntity):
    """
    Represents a not found object when creating an ``AuditLog``.
    
    Attributes
    ----------
    id : `int`
        The entity's unique identifier number.
    name : `str`
        The entity's name.
    type : `str`
        The entity's respective type's respective name.
        
        Can be one of:
        - `'Channel'`
        - `'Emoji'`
        - `'Integration'`
        - `'Invite'`
        - `'Message'`
        - `'Role'`
        - `'User'`
        - `'Webhook'`
        - `'Sticker'`
        - `'ScheduledEvent'`
        - `'ChannelThread'`
        - `'Integration'`
    """
    __slots__ = ('name', 'type', )
    
    def __init__(self, type_, id_, name=None):
        """
        Creates a new ``Unknown`` object from the given parameters.
        
        Parameters
        ----------
        type_ : `str`
            The entity's respective type's respective name.
        id_ : `int`
            The entity's unique identifier number.
        name : `str`, Optional
            The name of the entity if applicable. If not, `type_` will be used as name instead.
        """
        self.type = type_
        self.id = id_
        if name is None:
            name = type_
        self.name = name
    
    def __repr__(self):
        """Returns the representation of the entity."""
        return f'<{self.__class__.__name__} type={self.type} id={self.id} name={self.name!r}>'
    
    def __gt__(self, other):
        """Returns whether this entity's respective type matches with the other's and it's id is greater than the
        other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id > other.id)
    
    def __ge__(self ,other):
        """Returns whether this entity's respective type matches with the other's and it's id is greater or equal to
        the other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id >= other.id)
    
    def __eq__(self, other):
        """Returns whether this entity's respective type matches with the other's and it's id equals to the other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id == other.id)
    
    def __ne__(self,other):
        """Returns whether this entity's respective type matches with the other's and it's id not equals to the
        other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id != other.id)
    
    def __le__(self,other):
        """Returns whether this entity's respective type matches with the other's and it's id is less or equal to
        the other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id <= other.id)
    
    def __lt__(self,other):
        """Returns whether this entity's respective type matches with the other's and it's id is less than the
        other's."""
        if type(other) is type(self):
            if self.type != other.type:
                return NotImplemented
        elif isinstance(other, DiscordEntity):
            if self.type not in other.__class__.__name__:
                return NotImplemented
        else:
            return NotImplemented
        
        return (self.id < other.id)

class Gift:
    """
    Represents a Discord gift.
    
    Attributes
    ----------
    code : `str`
        The code of the gift.
    uses : `int`
        The amount how much time the gift can be used.
    """
    __slots__ = ('code', 'uses', )
    def __init__(self, data):
        """
        Creates a new ``Gift`` object from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Gift data received from Discord.
        """
        self.uses = data['uses']
        self.code = data['code']


def url_cutter(url):
    """
    Cuts down the given url to a more suitable length.
    
    Parameters
    ----------
    url : `str`
    
    Returns
    -------
    result : `str`
    """
    if len(url) < 50:
        return url
    
    position = url.find('/')
    
    if position == -1:
        return f'{url[:28]}...{url[-19:]}'
    
    position += 1
    if url[position] == '/':
        position += 1
        if position == len(url):
            return f'{url[:28]}...{url[-19:]}'
        
        position = url.find('/', position)
        position +=1
        if position == 0 or position == len(url):
            return f'{url[:28]}...{url[-19:]}'
    
    positions = [position]
    
    while True:
        position = url.find('/', position)
        if position == -1:
            break
        position +=1
        if position == len(url):
            break
        positions.append(position)
    
    from_start = 0
    from_end = 0
    top_limit = len(url)
    index = 0
    
    while True:
        value = positions[index]
        if value+from_end > 47:
            if from_start+from_end < 33:
                from_start = 47-from_end
                break
            else:
                index +=1
                if index == len(positions):
                    value  =0
                else:
                    value = positions[len(positions)-index]
                value = top_limit-value
                if value+from_start>47:
                    break
                else:
                    from_end = value
                    break
        from_start = value
        
        index = index+1
        value = positions[len(positions)-index]
        value = top_limit-value
        if value+from_start > 47:
            if from_start+from_end < 33:
                from_end = 47-from_start
                break
            else:
                if index == len(positions):
                    value = top_limit
                else:
                    value = positions[index]
                
                if value+from_end > 47:
                    break
                else:
                    from_start = value
                    break
        from_end = value
        
    return f'{url[:from_start]}...{url[top_limit-from_end-1:]}'

DELTA_RP = re_compile('([\+\-]?\d+)[ \t]*([a-zA-Z]+)')
TDELTA_KEYS = ('weeks', 'days', 'hours', 'minutes', 'seconds', 'microseconds')
RDELTA_KEYS = ('years', 'months', *TDELTA_KEYS)

def parse_tdelta(text):
    """
    Tries to parse out a ``timedelta`` from the inputted text.
    
    Returns
    -------
    tdelta : `None` or `datetime.timedelta`
    """
    text = text.lower()
    
    result = {}
    index = 0
    limit = len(TDELTA_KEYS)
    for amount, name in DELTA_RP.findall(text):
        if index == limit:
            break
        
        while True:
            key = TDELTA_KEYS[index]
            index += 1
            if key.startswith(name):
                result.setdefault(key, int(amount))
                break
            
            if index == limit:
                break
    
    if result:
        return timedelta(**result)

if relativedelta is None:
    parse_rdelta = None
else:
    def parse_rdelta(text):
        """
        Tries to parse out a ``relativedelta`` from the inputted text.
        
        Returns
        -------
        rdelta : `None` or `dateutil.relativedelta.relativedelta`
        """
        text = text.lower()
        
        result = {}
        index = 0
        limit = len(RDELTA_KEYS)
        for amount, name in DELTA_RP.findall(text):
            if index == limit:
                break
            
            while True:
                key = RDELTA_KEYS[index]
                index += 1
                if key.startswith(name):
                    result.setdefault(key, int(amount))
                    break
                
                if index == limit:
                    break
        
        if result:
            return relativedelta(**result)

CHANNEL_MESSAGE_RP = re_compile('(\d{7,21})-(\d{7,21})')

def parse_message_reference(text):
    """
    Tries the parse a ``Message``'s reference from the inputted text.
    
    Accepts the following formats:
    - `{message.id}`
    - `{channel.id}-{message.id}`
    - `{message.url}`
    
    Returns
    -------
    reference : `None` or `tuple` (`int`, `int`, `int`)
        On successful parsing returns a tuple of 3 elements:
        +-------------------+-------------------+-------------------------------+-------------------+---------------+
        | Respective name   | Parse-able from   | Parse-able from               | Parse-able from   | Default value |
        |                   | `{message.id}`    | `{channel.id}-{message.id}`   | `{message.url}`   |               |
        +===================+===================+===============================+=================+++===============+
        | `guild.id`        | False             | False                         | True              | `0`           |
        +-------------------+-------------------+-------------------------------+-------------------+---------------+
        | `channel.id       | False             | True                          | True              | `0`           |
        +-------------------+-------------------+-------------------------------+-------------------+---------------+
        | `message.id`      | True              | True                          | True              | N/A           |
        +-------------------+-------------------+-------------------------------+-------------------+---------------+
    """
    parsed = ID_RP.fullmatch(text)
    if (parsed is not None):
        message_id = int(parsed.group(1))
        
        guild_id = 0
        channel_id = 0
    else:
        parsed = CHANNEL_MESSAGE_RP.fullmatch(text)
        if (parsed is not None):
            channel_id, message_id = parsed.groups()
            channel_id = int(channel_id)
            message_id = int(message_id)
            
            guild_id = 0
        else:
            parsed = MESSAGE_JUMP_URL_RP.fullmatch(text)
            if (parsed is not None):
                guild_id, channel_id, message_id = parsed.groups()
                if guild_id == '@me':
                    guild_id = 0
                else:
                    guild_id = int(guild_id)
                channel_id = int(channel_id)
                message_id = int(message_id)
            else:
                return None
    
    return guild_id, channel_id, message_id


def sanitise_mention_escaper(transformations, match):
    """
    used inside of ``sanitize_mentions`` to escape mentions.
    
    Parameters
    ----------
    transformations : `dict` of (`str`, `str`) items
        Escape table.
    match : `re.Match`
        The matched mention to escape.
    
    Returns
    -------
    escaped : `str`
    """
    mention = match.group(0)
    return transformations.get(mention, mention)


def sanitize_mentions(content, guild=None):
    """
    Sanitizes the given content, removing the mentions from it.
    
    Parameters
    ----------
    content : `None` or `str`
        The content to sanitize.
    guild : `None` or ``Guild``, Optional
        Respective context to look up guild specific names of entities.
    
    Returns
    -------
    content : `None` or `str`
    """
    if (content is None):
        return
    
    transformations = {
        '@everyone':'@\u200beveryone',
        '@here':'@\u200bhere',
    }
    
    for id_ in USER_MENTION_RP.findall(content):
        id_ = int(id_)
        user = USERS.get(id_, None)
        if (user is None) or user.partial:
            sanitized_mention = '@invalid-user'
        else:
            sanitized_mention = '@'+user.name_at(guild)
        
        transformations[f'<@{id_}>'] = sanitized_mention
        transformations[f'<@!{id_}>'] = sanitized_mention
        
    for id_ in CHANNEL_MENTION_RP.findall(content):
        id_ = int(id_)
        channel = CHANNELS.get(id_, None)
        if (channel is None) or channel.partial:
            sanitized_mention = '@deleted channel'
        else:
            sanitized_mention = '#'+channel.name
        
        transformations[f'<#{id_}>'] = sanitized_mention
    
    for id_ in ROLE_MENTION_RP.findall(content):
        id_ = int(id_)
        role = ROLES.get(id_, None)
        if (role is None) or role.partial:
            sanitized_mention = '@deleted role'
        else:
            sanitized_mention = '@'+role.name
        
        transformations[f'<@&{id_}>'] = sanitized_mention
    
    return re_compile('|'.join(transformations)).sub(partial_func(sanitise_mention_escaper, transformations), content)


def sanitize_content(content, guild=None):
    """
    Sanitizes the markdown and the mentions in the given content.
    
    Parameters
    ----------
    content : `None` or `str`
        The content to sanitize.
    guild : `None` or ``Guild``, Optional
        Respective context to look up guild specific names of entities.
    
    Returns
    -------
    content : `None` or `str`
    """
    content = escape_markdown(content)
    content = sanitize_mentions(content, guild=guild)
    return content


def escape_markdown(content):
    """
    Escapes markdown from the given content.
    
    Parameters
    ----------
    content : `None` or `str`
        The content to sanitize.
    
    Returns
    -------
    content : `None` or `str`
    """
    if (content is None):
        return
    
    content = content.replace('\\', '\\\\')
    content = content.replace('_', '\\_')
    content = content.replace('*', '\\*')
    content = content.replace('|', '\\|')
    content = content.replace('~', '\\~')
    content = content.replace('`', '\\`')
    content = content.replace('>', '\\>')
    content = content.replace(':', '\\:')
    return content


def parse_date_header_to_datetime(date_data):
    """
    Parsers header date value to `datetime`.
    
    Parameters
    ----------
    date_data : ``str``
        Date value inside of a header.

    Returns
    -------
    date : `datetime`
        The parsed out date time.
    """
    *date_tuple, tz = parse_date_timezone(date_data)
    if tz is None:
        date = datetime(*date_tuple[:6])
    else:
        date = datetime(*date_tuple[:6], tzinfo=timezone(timedelta(seconds=tz)))
    return date


URL_RP = re_compile(
    # protocol identifier
    '(?:(?:https?|ftp)://)'
    # user:pass authentication
    '(?:[-a-z\u00a1-\uffff0-9._~%!$&\'()*+,;=:]+'
    '(?::[-a-z0-9._~%!$&\'()*+,;=:]*)?@)?'
    '(?:'
    '(?:'
    # IP address exclusion
    # private & local networks
    '(?:(?:10|127)(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}(?:\.(?:0|[1-9]\d?|1\d\d|2[0-4]\d|25[0-5])))|'
    '(?:(?:169\.254|192\.168)(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5]))(?:\.(?:0|[1-9]\d?|1\d\d|2[0-4]\d|25[0-5])))|'
    '(?:172\.(?:1[6-9]|2\d|3[0-1])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5]))(?:\.(?:0|[1-9]\d?|1\d\d|2[0-4]\d|25[0-5]))))'
    '|'
    # private & local hosts
    '(?:localhost)'
    '|'
    # IP address dotted notation octets
    # excludes loop back network 0.0.0.0
    # excludes reserved space >= 224.0.0.0
    # excludes network & broadcast addresses
    # (first & last IP address of each class)
    '(?:'
    '(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])'
    '(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}'
    '(?:\.(?:0|[1-9]\d?|1\d\d|2[0-4]\d|25[0-5])))'
    '|'
    # IPv6 RegEx
    '\[('
    # 1:2:3:4:5:6:7:8
    '([0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}|'
    # 1::                              1:2:3:4:5:6:7::
    '([0-9a-fA-F]{1,4}:){1,7}:|'
    # 1::8             1:2:3:4:5:6::8  1:2:3:4:5:6::8
    '([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|'
    # 1::7:8           1:2:3:4:5::7:8  1:2:3:4:5::8
    '([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|'
    # 1::6:7:8         1:2:3:4::6:7:8  1:2:3:4::8
    '([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|'
    # 1::5:6:7:8       1:2:3::5:6:7:8  1:2:3::8
    '([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|'
    # 1::4:5:6:7:8     1:2::4:5:6:7:8  1:2::8
    '([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|'
    # 1::3:4:5:6:7:8   1::3:4:5:6:7:8  1::8
    '[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|'
    # ::2:3:4:5:6:7:8  ::2:3:4:5:6:7:8 ::8       ::
    ':((:[0-9a-fA-F]{1,4}){1,7}|:)|'
    # fe80::7:8%eth0   fe80::7:8%1
    # (link-local IPv6 addresses with zone index)
    'fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]?|::(ffff(:0{1,4})?:)?((25[0-5]|(2[0-4]|1?[0-9])?[0-9])\.){3}'
    # ::255.255.255.255   ::ffff:255.255.255.255  ::ffff:0:255.255.255.255
    # (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
    '(25[0-5]|(2[0-4]|1?[0-9])?[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1?[0-9])?[0-9])\.){3}'
    # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33
    # (IPv4-Embedded IPv6 Address)
    '(25[0-5]|(2[0-4]|1?[0-9])?[0-9])'
    ')\]|'
    # host name
    '(?:(?:(?:xn--)|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)'
    # domain name
    '(?:\.(?:(?:xn--)|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)*'
    # TLD identifier
    '(?:\.(?:(?:xn--[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]{2,})|[a-z\u00a1-\uffff\U00010000-\U0010ffff]{2,}))'
    ')'
    # port number
    '(?::\d{2,5})?'
    # resource path
    '(?:/[-a-z\u00a1-\uffff\U00010000-\U0010ffff0-9._~%!$&\'()*+,;=:@/]*)?'
    # query string
    '(?:\?\S*)?'
    # fragment
    '(?:#\S*)?',
    re_unicode|re_ignore_case
)

def is_url(url):
    """
    Returns whether the given value is url.
    
    Parameters
    ----------
    url : `str`
        The url to validate.
    
    Returns
    -------
    is_url : `bool`
    """
    return (URL_RP.fullmatch(url) is not None)


@modulize
class TIMESTAMP_STYLES:
    """
    Contains timestamp format styles, which can be used within Discord's markdown format.
    You may be use these styles at ``format_datetime``, ``format_id`` and at ``format_unix_time``.
    
    The style formats are the following:
    
    +-------------------+-------+-----------+
    | Name              | Value | Note      |
    +====================+======+===========+
    | short_time        | `'t'` |           |
    +-------------------+-------+-----------+
    | long_time         | `'T'` |           |
    +-------------------+-------+-----------+
    | short_date        | `'d'` |           |
    +-------------------+-------+-----------+
    | long_date         | `'D'` |           |
    +-------------------+-------+-----------+
    | short_date_time   | `'f'` | default   |
    +-------------------+-------+-----------+
    | long_date_time    | `'F'` |           |
    +-------------------+-------+-----------+
    | relative_time     | `'R'` |           |
    +-------------------+-------+-----------+
    
    Note, that Discord's time formatting is localized and they are all stultus when english language is selected.
    To avoid insanity, I beg you to use
    `datetime formatting:https://docs.python.org/3/library/datetime.html#datetime.date.__format__` instead.
    
    > "wen day is dark always rember happy day"
    
    As a quick example, hata internally uses the following formatting:
    
    ```py
    >>> from hata import DATETIME_FORMAT_CODE
    >>> from datetime import datetime
    >>> print(f'{datetime.utcnow():{DATETIME_FORMAT_CODE}}')
    2021-08-05 13:53:16
    ```
    
    Hata also gives option for relative formatting with the ``elapsed_time`` function. It is available when
    `dateutil:https://pypi.org/project/python-dateutil/` is installed.
    
    ```py
    >>> from hata import elapsed_time
    >>> from datetime import datetime, timedelta
    >>> when = datetime.utcnow()-timedelta(days=5)
    >>> print(f'{elapsed_time(when)} ago')
    5 days ago
    ```
    """
    short_time = 't'
    long_time = 'T'
    short_date = 'd'
    long_date = 'D'
    short_date_time = 'f'
    long_date_time = 'F'
    relative_time = 'R'


def format_datetime(date_time, style=None):
    """
    Formats date time to Discord's timestamp markdown format.
    
    For formatting details please check out ``TIMESTAMP_STYLES``, which contains the usable styles.
    
    Parameters
    ----------
    date_time : `datetime`
        The datetime to format.
    style : `None` or `str`, `optional
        Format code to use. They are listed within ``TIMESTAMP_STYLES``.
    
    Returns
    -------
    formatted_string : `str`
    """
    return format_unix_time(datetime_to_unix_time(date_time), style)


def format_id(id_, style=None):
    """
    Formats Discord identifier to Discord's timestamp markdown format.
    
    For formatting details please check out ``TIMESTAMP_STYLES``, which contains the usable styles.
    
    Parameters
    ----------
    id_ : `int`
        The Discord identifier to format.
    style : `None` or `str`, `optional
        Format code to use. They are listed within ``TIMESTAMP_STYLES``.
    
    Returns
    -------
    formatted_string : `str`
    """
    return format_unix_time(id_to_unix_time(id_), style)


def format_unix_time(unix_time, style=None):
    """
    Formats unix time to Discord'stimestamp  markdown format.
    
    For formatting details please check out ``TIMESTAMP_STYLES``, which contains the usable styles.
    
    Parameters
    ----------
    unix_time : `int`
        The datetime to format.
    style : `None` or `str`, `optional
        Format code to use. They are listed within ``TIMESTAMP_STYLES``.
    
    Returns
    -------
    formatted_string : `str`
    """
    if style is None:
        formatted_string = f'<t:{unix_time}>'
    else:
        formatted_string = f'<t:{unix_time}:{style}>'
    
    return formatted_string



__all__ = ()

from time import time as time_now

try:
    import nacl.secret
except ImportError:
    SecretBox = None
else:
    SecretBox = nacl.secret.SecretBox
    del nacl

from ...backend.exceptions import ConnectionClosed
from ...backend.utils import to_json, from_json

from ..exceptions import VOICE_CLIENT_DISCONNECT_CLOSE_CODE

from .heartbeat import Kokoro
from .rate_limit import GatewayRateLimiter


IDENTIFY = 0
SELECT_PROTOCOL = 1
READY = 2
HEARTBEAT = 3
SESSION_DESCRIPTION = 4
SPEAKING = 5
HEARTBEAT_ACK = 6
RESUME = 7
HELLO = 8
INVALIDATE_SESSION = 9
CLIENT_CONNECT = 12
CLIENT_DISCONNECT = 13
VIDEO_SESSION_DESCRIPTION = 14
VIDEO_SINK = 15

"""
IDENTIFY : `int` = `0`
    Send only, used at ``._identify``.
SELECT_PROTOCOL : `int` = `1`
    Send only, used at ``._select_protocol``.
READY : `int` = `2`
    Receive only, used at ``._initial_connection``.
HEARTBEAT : `int` = `3`
    Send only, used at ``._beat``.
SESSION_DESCRIPTION : `int` = `4`
    Receive only, used at ``._received_message``.
SPEAKING : `int` = `5`
    Send and receive, used at ``._set_speaking`` and at ``._received_message``.
HEARTBEAT_ACK : `int` = `6`
    Receive only, used at ``._received_message``.
RESUME : `int` = `7`
    Send only, used at ``._resume``.
HELLO : `int` = `8`
    Receive only, used at ``._received_message``.
INVALIDATE_SESSION : `int` = `9`
    Receive only, used at ``._received_message``.
CLIENT_CONNECT : `int` = `12`
    Receive only, used at ``._received_message``.
CLIENT_DISCONNECT : `int` = `13`
    Receive only, used at ``._received_message``.
VIDEO_SESSION_DESCRIPTION : `int` = `14`
    Receive only. Not used.
VIDEO_SINK : `int` = `15`
    Receive and send, not used.
"""

class DiscordGatewayVoice:
    """
    The gateway used by ``VoiceClient``-s to communicate with Discord with secure websocket.
    
    Attributes
    ----------
    client : ``VoiceClient``
        The owner voice client of the gateway.
    kokoro : `None` or `Kokoro`
        The heart of the gateway, sends beat-data at set intervals. If does not receives answer in time, restarts
        the gateway.
    rate_limit_handler : ``GatewayRateLimiter``
        The rate limit handler of the gateway.
    websocket : `None` or `WSClient`
        The websocket client of the gateway.
    """
    __slots__ = ('client', 'kokoro', 'rate_limit_handler', 'websocket')
    
    def __init__(self, voice_client):
        """
        Creates a voice gateway with it's default attributes.
        
        Parameters
        ----------
        client : ``VoiceClient``
            The owner client of the gateway.
        """
        self.websocket = None
        self.client = voice_client
        
        self.kokoro = None
        self.rate_limit_handler = GatewayRateLimiter()
    
    
    async def start(self):
        """
        Starts the gateway's ``.kokoro``.
        
        This method is a coroutine.
        """
        kokoro = self.kokoro
        if kokoro is None:
            self.kokoro = await Kokoro(self)
        else:
            await kokoro.restart()
    
    # connecting, message receive and processing
    
    async def connect(self, resume=False):
        """
        Connects the gateway to Discord. If the connecting was successful, will start it's `.kokoro` as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        resume : `bool`
            Whether the gateway should try to resume the existing connection.
        
        Raises
        ------
        ConnectionError
        OSError
        ValueError
        ConnectionClosed
        InvalidHandshake
        WebSocketProtocolError
        """
        kokoro = self.kokoro
        if (kokoro is not None):
            kokoro.terminate()
            del kokoro
        
        websocket = self.websocket
        if (websocket is not None) and (not websocket.closed):
            await websocket.close(4000)
            self.websocket = None
        
        gateway = f'wss://{self.client._endpoint}/?v=4'
        self.websocket = await self.client.client.http.connect_ws(gateway)
        
        kokoro = self.kokoro
        if kokoro is None:
            self.kokoro = kokoro = await Kokoro(self)
        kokoro.start_beating()
        del kokoro
        
        if resume:
            await self._resume()
        else:
            await self._identify()
    
    
    async def _poll_event(self):
        """
        Waits to receive a message from Discord, then calls ``._received_message``.
        
        This method is a coroutine.
        
        Raises
        ------
        TimeoutError
            If the gateways's `.kokoro` is not beating, meanwhile it should.
        ConnectionClosed
            If the websocket is already closed. Can happen when destroying ghost clients.
        """
        websocket = self.websocket
        if websocket is None:
            raise ConnectionClosed(VOICE_CLIENT_DISCONNECT_CLOSE_CODE, None)
        
        message = await websocket.receive()
        await self._received_message(message)
    
    
    async def _received_message(self, message):
        """
        Processes the message sent by Discord.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : `bytes`
            The received message.
        
        Raises
        ------
        TimeoutError
            If the gateways's `.kokoro` is not beating, meanwhile it should.
        """
        message = from_json(message)
        
        operation = message['op']
        try:
            data = message['d']
        except KeyError:
            data = None
        
        if operation == CLIENT_CONNECT:
            user_id = int(data['user_id'])
            try:
                audio_source = data['audio_ssrc']
            except KeyError:
                pass
            else:
                self.client._update_audio_source(user_id, audio_source)
            
            try:
                video_source = data['video_ssrc']
            except KeyError:
                pass
            else:
                self.client._update_video_source(user_id, video_source)
            
            return
        
        if operation == SPEAKING:
            user_id = int(data['user_id'])
            audio_source = data['ssrc']
            self.client._update_audio_source(user_id, audio_source)
            return
        
        if operation == CLIENT_DISCONNECT:
            user_id = int(data['user_id'])
            self.client._remove_source(user_id)
            return
        
        kokoro = self.kokoro
        if kokoro is None:
            kokoro = await Kokoro(self)
        
        if operation == HELLO:
            # sowwy, but we need to ignore these or we will keep getting timeout
            # kokoro.interval=data['heartbeat_interval']/100.
            # send a heartbeat immediately
            await kokoro.beat_now()
            return
        
        if operation == HEARTBEAT_ACK:
            kokoro.answered()
            return
        
        if operation == SESSION_DESCRIPTION:
            # data['mode'] is same as our default every time?
            self.client._secret_box = SecretBox(bytes(data['secret_key']))
            if kokoro.beater is None: # Discord order bug ?
                kokoro.start_beating()
                await self._beat()
            
            await self._set_speaking(self.client.speaking)
            return
        
        if kokoro.beater is None:
            raise TimeoutError
        
        if operation == READY:
            # need to ignore interval
            # kokoro.interval = data['heartbeat_interval']/100.
            await self._initial_connection(data)
            return
        
        if operation == INVALIDATE_SESSION:
            await self._identify()
            return
        
        # Ignore VIDEO_SESSION_DESCRIPTION and VIDEO_SINK for now
    
    # general stuffs
    
    @property
    def latency(self):
        """
        The latency of the websocket in seconds. If no latency is recorded, will return `Kokoro.DEFAULT_LATENCY`.
        
        Returns
        -------
        latency : `float`
        """
        kokoro = self.kokoro
        if kokoro is None:
            return Kokoro.DEFAULT_LATENCY
        return kokoro.latency
    
    
    async def terminate(self):
        """
        Terminates the gateway's ``.kokoro`` and closes it's `.websocket`` with close code of `4000`.
        
        This method is a coroutine.
        """
        kokoro = self.kokoro
        if kokoro is not None:
            kokoro.terminate()
            
        websocket = self.websocket
        if websocket is None:
            return
        self.websocket = None
        await websocket.close(4000)
    
    
    async def close(self):
        """
        Cancels the gateway's ``.kokoro`` and closes it's ``.websocket`` with close code of `1000`.
        
        This method is a coroutine.
        """
        self.rate_limit_handler.cancel()
        
        kokoro = self.kokoro
        if (kokoro is not None):
            self.kokoro = None
            kokoro.cancel()
        
        websocket = self.websocket
        if websocket is None:
            return
        
        self.websocket = None
        await websocket.close(1000)
    
    
    async def send_as_json(self, data):
        """
        Sends the data as json to Discord on the gateway's ``.websocket``. If there is no websocket, or the websocket
        is closed will not raise.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items or `list` of `Any`
        """
        websocket = self.websocket
        if websocket is None:
            return
        
        if await self.rate_limit_handler:
            return
        
        try:
            await websocket.send(to_json(data))
        except ConnectionClosed:
            pass
    
    
    def __repr__(self):
        """Returns the representation of the gateway."""
        return f'<{self.__class__.__name__} of {self.client}>'
    
    #special operations
    
    async def _identify(self):
        """
        Sends an `IDENTIFY` packet to Discord.
        
        This method is a coroutine.
        """
        voice_client = self.client
        
        data = {
            'op': IDENTIFY,
            'd': {
                'server_id': str(voice_client.guild_id),
                'user_id': str(voice_client.client.id),
                'session_id': voice_client._session_id,
                'token': voice_client._token,
            },
        }
        
        await self.send_as_json(data)
    
    
    async def _resume(self):
        """
        Sends a `RESUME` packet to Discord.
        
        This method is a coroutine.
        """
        voice_client = self.client
        
        data = {
            'op': RESUME,
            'd': {
                'token': voice_client._token,
                'server_id': str(voice_client.guidld_id),
                'session_id': voice_client._session_id,
            },
        }
        
        await self.send_as_json(data)
    
    
    async def _select_protocol(self, ip, port):
        """
        Sends a `SELECT_PROTOCOL` packet to Discord.
        
        This method is a coroutine.
        
        Parameters
        ----------
        ip : `str`
            The received ip of the voice client to use.
        port : `int`
            The received port of the voice client to use.
        """
        data = {
            'op': SELECT_PROTOCOL,
            'd': {
                'protocol': 'udp',
                'data': {
                    'address': ip,
                    'port': port,
                    'mode': 'xsalsa20_poly1305'
                },
            },
        }
        
        await self.send_as_json(data)
    
    
    async def _beat(self):
        """
        Sends a `HEARTBEAT` packet to Discord.
        
        This method is a coroutine.
        """
        data = {
            'op': HEARTBEAT,
            'd': int(time_now()*1000),
        }
        
        await self.send_as_json(data)
    
    
    async def _initial_connection(self, data):
        """
        Processes the data from `READY` operation and selects protocol with ``._select_protocol``.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received data from Discord.
        """
        voice_client = self.client
        voice_client._audio_source = data['ssrc']
        voice_client._audio_port = data['port']
        voice_client._endpoint_ip = data['ip']
        packet = bytearray(70)
        packet[0:4] = voice_client._audio_source.to_bytes(4, 'big')
        
        voice_client.send_packet(packet)
        
        protocol = voice_client._protocol
        # Make sure, that the voice client's reader do not wanna read our data away from us.
        protocol.cancel_current_reader()
        received = await protocol.read(70)
        
        # the ip is ascii starting at the 4th byte and ending at the first null
        voice_client._ip = ip = received[4:received.index(0, 4)].decode('ascii')
        voice_client._port = port = int.from_bytes(received[-2:], 'big')
        
        await self._select_protocol(ip, port)
        await self._client_connect()
   
    
    async def _client_connect(self):
        """
        Sends a `CLIENT_CONNECT` packet to Discord.
        
        This method is a coroutine.
        """
        voice_client = self.client
        data = {
            'op': CLIENT_CONNECT,
            'd': {
                'audio_ssrc': voice_client._audio_source,
                'video_ssrc': voice_client._video_source,
                'rtx_ssrc': 0,
            }
        }
        
        await self.send_as_json(data)
    
    
    async def _set_speaking(self, is_speaking):
        """
        Sends a `SPEAKING` packet with given `is_speaking` state.
        
        This method is a coroutine.
        
        Parameters
        ----------
        is_speaking : `bool`
            Whether the voice client should show up as speaking and be able to send voice data or not.
        """
        data = {
            'op':SPEAKING,
            'd': {
                'speaking': is_speaking,
                'delay': 0,
            },
        }
        
        await self.send_as_json(data)



__all__ = ('Webhook', )

from ...backend.export import export, include
from ...backend.utils import copy_docs

from ..core import USERS
from ..user import User, ZEROUSER, ClientUserBase, UserBase
from ..user.helpers import get_banner_color_from_data
from ..exceptions import DiscordException, ERROR_CODES
from ..preconverters import preconvert_snowflake, preconvert_str, preconvert_preinstanced_type
from ..bases import ICON_TYPE_NONE, Icon, instance_or_id_to_instance
from ..http.urls import WEBHOOK_URL_PATTERN
from ..http import urls as module_urls
from ..color import Color

from .webhook_base import WebhookBase
from .preinstanced import WebhookType
from .webhook_sources import WebhookSourceChannel, WebhookSourceGuild

create_partial_webhook_from_id = include('create_partial_webhook_from_id')
create_partial_channel_from_id = include('create_partial_channel_from_id')
ChannelText = include('ChannelText')
Client = include('Client')


@export
class Webhook(WebhookBase):
    """
    Represents a Discord webhook. At some cases it might be used as webhook's user representation.
    
    Attributes
    ----------
    id : `int`
        The webhook's unique identifier number.
    name : str
        The webhook's username.
    discriminator : `int`
        The webhook's discriminator. Given to avoid overlapping names.
    avatar_hash : `int`
        The webhook's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The webhook's avatar's type.
    banner_color : `None` or ``Color``
        The user's banner color if has any.
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    banner_type : ``IconType``
        The user's banner's type.
    channel_id : `int`
        The channel's identifier, where the webhook is going to send it's messages.
    type : ``WebhookType``
        The webhook's type.
    application_id : `int`
        The application's id what created the webhook. Defaults to `0` if not applicable.
    source_channel : `None` or ``WebhookSourceChannel``
        The webhook's source channel. Applicable for ``WebhookType.server`` webhooks.
    source_guild : `None` or ``WebhookSourceGuild``
        The webhook's source guild. Applicable for ``WebhookType.server`` webhooks.
    token : `str`
        The webhooks's token. You need an `id` and a `token` to send webhook message. Defaults to empty string.
    user : ``ClientUserBase``
        The creator of the webhook, or `ZEROUSER` if unknown.
    
    Notes
    -----
    Instances of this class are weakreferable.
    """
    __slots__ = ('application_id', 'source_channel', 'source_guild', 'token', 'user')
    
    def __new__(cls, data):
        """
        Tries to get the webhook from the existing ones, then update it. If no webhook was found, creates a new one and
        fills it's attributes from the data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received webhook data.
        
        Returns
        -------
        webhook : ``Webhook``
        """
        webhook_id = int(data['id'])
        try:
            self = USERS[webhook_id]
        except KeyError:
            self = object.__new__(cls)
            USERS[webhook_id] = self
            self.id = webhook_id
            self.token = ''
        
        self._set_attributes(data)
        self.type = WebhookType.get(data['type'])
        
        application_id = data.get('application_id', None)
        if application_id is None:
            application_id = 0
        else:
            application_id = int(application_id)
        self.application_id = application_id
        
        source_channel_data = data.get('source_channel', None)
        if source_channel_data is None:
            source_channel = None
        else:
            source_channel = WebhookSourceChannel(source_channel_data)
        self.source_channel = source_channel
        
        source_guild_data = data.get('source_guild', None)
        if source_guild_data is None:
            source_guild = None
        else:
            source_guild = WebhookSourceGuild(source_guild_data)
        self.source_guild = source_guild
        
        return self
    
    
    @classmethod
    def from_url(cls, url):
        """
        Tries to parse the webhook's `id` and `token` from the given `url`. If succeeds, returns a partial webhook.
        If parsing fails, returns `None`.
        
        Parameters
        ----------
        url : `str`
            The url of the webhook.
        
        Returns
        -------
        webhook : `None` or ``Webhook``
        """
        result = WEBHOOK_URL_PATTERN.fullmatch(url)
        if result is None:
            return None
        
        webhook_id = int(result.group(1))
        webhook_token = result.group(2)
        
        return create_partial_webhook_from_id(webhook_id, webhook_token)
    
    
    def _set_attributes(self, data):
        self._update_attributes(data)
        
        self.channel_id = int(data['channel_id'])
        
        token = data.get('token', None)
        if (token is not None):
            self.token = token
        
        try:
            user_data = data['user']
        except KeyError:
            user = ZEROUSER
        else:
            user = User(user_data)
        self.user = user
    
    
    @classmethod
    def precreate(cls, webhook_id, **kwargs):
        """
        Precreates the webhook by creating a partial one with the given parameters. When the webhook will be loaded,
        the precreated one will be picked up and will be updated. If an already existing webhook would be precreated,
        it will be updated with the given parameters only if it is a partial one.
        
        Parameters
        ----------
        webhook_id : `int` or `str`
            The webhook's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the webhook.
        
        Other Parameters
        ----------------
        name : `str`, Optional (Keyword only)
            The webhook's ``.name``.
        token : `str`, Optional (Keyword only)
            The webhook's ``.token``.
        
        avatar : `None`, ``Icon`` or `str`, Optional (Keyword only)
            The webhook's avatar.
            
            > Mutually exclusive with `avatar_type` and `avatar_hash`.
        
        avatar_type : ``IconType``, Optional (Keyword only)
            The webhook's avatar's type.
            
            > Mutually exclusive with `avatar_type`.
        
        avatar_hash : `int`, Optional (Keyword only)
            The webhook's avatar's hash.
            
            > Mutually exclusive with `avatar`.
        
        banner : `None`, ``Icon`` or `str`, Optional (Keyword only)
            The webhook's banner.
            
            > Mutually exclusive with `banner_type` and `banner_hash`.
        
        banner_type : ``IconType``, Optional (Keyword only)
            The webhook's banner's type.
            
            > Mutually exclusive with `banner_type`.
        
        banner_hash : `int`, Optional (Keyword only)
            The webhook's banner hash.
            
            > Mutually exclusive with `banner`.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The webhook's ``.user``.
        channel : ``ChannelText``, Optional (Keyword only)
            The webhook's ``.channel``.
        application_id : `int`, Optional (Keyword only)
            The application's id what created the webhook.
        
        Returns
        -------
        webhook : ``Webhook``
        """
        webhook_id = preconvert_snowflake(webhook_id, 'webhook_id')
        
        if kwargs:
            processable = []
            
            for attribute_name, lower_limit, upper_limit in (
                    ('name' , 0 , 80,),
                    ('token', 60, 68,),
                        ):
                
                try:
                    attribute_value = kwargs.pop(attribute_name)
                except KeyError:
                    pass
                else:
                    attribute_value = preconvert_str(attribute_value, attribute_name, lower_limit, upper_limit)
                    processable.append((attribute_name, attribute_value))
            
            cls.avatar.precovert(kwargs, processable)
            cls.banner.precovert(kwargs, processable)
            
            for attribute_name, attribute_type in (
                    ('user'      , (User, Client),),
                    ('channel_id', ChannelText   ,),
                        ):
                try:
                    attribute_value = kwargs.pop(attribute_name)
                except KeyError:
                    pass
                else:
                    attribute_value = instance_or_id_to_instance(attribute_value, attribute_type, attribute_name)
                    processable.append((attribute_name, attribute_value))
            
            try:
                type_ = kwargs.pop('type')
            except KeyError:
                pass
            else:
                type_ = preconvert_preinstanced_type(type_, 'type', WebhookType)
                processable.append(('type', type_))
            
            try:
                application_id = kwargs.pop('application_id')
            except KeyError:
                pass
            else:
                application_id = preconvert_snowflake(application_id, 'application_id')
                processable.append(('application_id', application_id))
        
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}')
        
        else:
            processable = None
        
        try:
            self = USERS[webhook_id]
        except KeyError:
            self = cls.create_empty(webhook_id)
            USERS[webhook_id] = self
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self
    
    
    url = property(module_urls.webhook_url)
    
    
    @classmethod
    async def _from_follow_data(cls, data, source_channel, target_channel_id, client):
        """
        Creates the webhook, what executes cross-posts.
        
        This method is ensured after following a channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received webhook data.
        source_channel : ``ChannelText`` instance
            The followed channel.
        target_channel_id : `int`
            The target channel's identifier where the webhook messages will be sent.
        client : ``Client``
            The client who created the webhook.
        
        Returns
        -------
        webhook : ``Webhook``
        """
        webhook_id = int(data['webhook_id'])
        
        guild = source_channel.guild
        if guild is None:
            try:
                extra_data = await client.http.webhook_get(webhook_id)
            except DiscordException as err:
                if err.code == ERROR_CODES.unknown_webhook:
                    #not lucky
                    name = ''
                    avatar_type = ICON_TYPE_NONE
                    avatar_hash = 0
                else:
                    raise
            
            else:
                name = extra_data['name']
                if name is None:
                    name = ''
                
                avatar_type, avatar_hash = Icon.from_base16_hash(data.get('avatar', None))
            
            source_guild_data = data.get('source_guild', None)
            if source_guild_data is None:
                source_guild = None
            else:
                source_guild = WebhookSourceGuild(source_guild_data)
        
        else:
            # TODO: can it be animated if the guild's icon is animated?
            avatar_hash = guild.icon_hash
            avatar_type = guild.icon_type
            name = f'{guild.name} #{source_channel.name}'
            
            source_guild = WebhookSourceGuild._from_guild(guild)
        
        source_channel = WebhookSourceChannel._from_channel(source_channel)
        
        self = object.__new__(cls)
        self.id = webhook_id

        self.discriminator = 0
        self.avatar_hash = avatar_hash
        self.avatar_type = avatar_type
        self.banner_color = None
        self.banner_hash = 0
        self.banner_type = ICON_TYPE_NONE
        self.name = name
        
        self.channel_id = target_channel_id
        self.type = WebhookType.server
        
        self.application_id = 0
        self.source_channel = source_channel
        self.source_guild = source_guild
        self.token = ''
        self.user = client
        
        USERS[webhook_id] = self
        
        return self

    @copy_docs(UserBase._set_default_attributes)
    def _set_default_attributes(self):
        WebhookBase._set_default_attributes(self)
        
        self.application_id = 0
        self.source_guild = None
        self.source_channel = None
        self.token = ''
        self.user = ZEROUSER



# -*- coding: utf-8 -*-
__all__ = ('KOKORO_STRATEGY', )

from threading import current_thread

from sqlalchemy import util
from sqlalchemy.engine import Engine
from sqlalchemy.engine.strategies import DefaultEngineStrategy

from ...backend.utils import alchemy_incendiary
from ...backend.executor import ExecutorThread


class KOKOROEngine:
    __slots__=('_engine', '_worker',)
    def __init__(self, pool, dialect, u, single_worker=True, **kwargs):
        if single_worker:
            worker = ExecutorThread()
        else:
            worker = None
        self._worker = worker
        self._engine = Engine(pool, dialect, u, **kwargs)
    
    @property
    def uses_single_worker(self):
        """
        Returns whether the engine uses a single worker.
        
        Returns
        -------
        uses_single_worker : `bool`
        """
        return (self._worker is not None)
    
    @property
    def dialect(self):
        return self._engine.dialect
    
    @property
    def _has_events(self):
        return self._engine._has_events
    
    @property
    def logger(self):
        return self._engine.logger
    
    @property
    def _execution_options(self):
        return self._engine._execution_options
    
    def _should_log_info(self):
        return self._engine._should_log_info()
    
    def connect(self):
        return ConnectionCM(self._connect())
    
    async def _connect(self):
        executor = self._worker
        if executor is None:
            executor = current_thread().claim_executor()
        connection = await executor.execute(self._engine.connect)
        return AsyncConnection(connection, executor)
    
    def begin(self, close_with_result=False):
        executor=self._worker
        if executor is None:
            executor = current_thread().claim_executor()
        return EngineTransactionCM(self, close_with_result, executor)
    
    async def execute(self, *args, **kwargs):
        executor = self._worker
        if executor is None:
            executor = current_thread().claim_executor()
        result_proxy = await executor.execute(alchemy_incendiary(self._engine.execute, args, kwargs))
        return AsyncResultProxy(result_proxy, executor)
    
    async def scalar(self, *args, **kwargs):
        executor = self._worker
        if executor is None:
            executor = current_thread().claim_executor()
        result_proxy = await executor.execute(alchemy_incendiary(self._engine.execute, args, kwargs))
        async_result_proxy = AsyncResultProxy(result_proxy, executor)
        return await async_result_proxy.scalar()
    
    async def has_table(self, table_name, schema=None):
        return await current_thread().run_in_executor(alchemy_incendiary(self._engine.has_table, (table_name, schema)))
    
    async def table_names(self, schema=None, connection=None):
        task = alchemy_incendiary(self._engine.table_names, (schema, None if connection else connection._connection))
        executor = self._worker
        if executor is None:
            return await current_thread().run_in_executor(task)
        else:
            return await executor.execute(task)
    
    def __del__(self):
        worker = self._worker
        if (worker is not None):
            self._worker = None
            worker.cancel()


class AsyncConnection:
    __slots__=('_connection', 'executor',)
    def __init__(self, connection, executor):
        self._connection = connection
        self.executor = executor
    
    async def execute(self, *args, **kwargs):
        result_proxy = await self.executor.execute(alchemy_incendiary(self._connection.execute, args, kwargs))
        return AsyncResultProxy(result_proxy, self.executor)
    
    async def scalar(self, *args, **kwargs):
        result_proxy = await self.executor.execute(alchemy_incendiary(self._connection.execute, args, kwargs))
        async_result_proxy = AsyncResultProxy(result_proxy, self.executor)
        return await async_result_proxy.scalar()
    
    async def close(self, *args, **kwargs):
        await self.executor.execute(alchemy_incendiary(self._connection.close, args, kwargs))
    
    @property
    def closed(self):
        return self._connection.closed
    
    def begin(self):
        return TransactionCM(self._begin())
    
    async def _begin(self):
        transaction = await self.executor.execute(self._connection.begin)
        return AsyncTransaction(transaction, self.executor)
    
    def begin_nested(self):
        return TransactionCM(self._begin_nested())
    
    async def _begin_nested(self):
        transaction = await self.executor.execute(self._connection.begin_nested)
        return AsyncTransaction(transaction, self.executor)
    
    def in_transaction(self):
        return self._connection.in_transaction()


class AsyncTransaction:
    __slots__ = ('_transaction', 'executor',)
    def __init__(self, transaction, executor):
        self._transaction = transaction
        self.executor = executor
    
    async def commit(self):
        return await self.executor.execute(self._transaction.commit)
    
    async def rollback(self):
        return await self.executor.execute(self._transaction.rollback)
    
    async def close(self):
        await self.executor.execute(self._transaction.close)


class AsyncResultProxyIterator:
    __slots__ = ('_result_proxy', 'executor',)
    def __init__(self, result_proxy, executor):
        self._result_proxy = result_proxy
        self.executor = executor
    
    def __aiter__(self):
        return self
    
    async def __anext__(self):
        row = await self.executor.execute(self._result_proxy.fetchone)
        if row is None:
            raise StopAsyncIteration
        return row


class AsyncResultProxy:
    __slots__ = ('_result_proxy', 'executor',)
    def __init__(self, result_proxy, executor):
        self._result_proxy = result_proxy
        self.executor = executor
    
    def __aiter__(self):
        return AsyncResultProxyIterator(self._result_proxy, self.executor)
    
    async def fetchone(self):
        return await self.executor.execute(self._result_proxy.fetchone)
    
    async def fetchmany(self, size=None):
        return await self.executor.execute(self._result_proxy.fetchmany, size=size)
    
    async def fetchall(self):
        return await self.executor.execute(self._result_proxy.fetchall)
    
    async def scalar(self):
        return await self.executor.execute(self._result_proxy.scalar)
    
    async def first(self):
        return await self.executor.execute(self._result_proxy.first)
    
    async def keys(self):
        return await self.executor.execute(self._result_proxy.keys)
    
    async def close(self):
        return await self.executor.execute(self._result_proxy.close)
    
    @property
    def returns_rows(self):
        return self._result_proxy.returns_rows
    
    @property
    def rowcount(self):
        return self._result_proxy.rowcount
    
    @property
    def inserted_primary_key(self):
        return self._result_proxy.inserted_primary_key


class EngineTransactionCM:
    __slots__ = ('_close_with_result', '_context', '_engine', 'executor',)

    def __init__(self, engine, close_with_result, executor):
        self._engine = engine
        self._close_with_result = close_with_result
        self.executor = executor
    
    async def __aenter__(self):
        self._context = await self.executor.execute(
            alchemy_incendiary(self._engine._engine.begin, (self._close_with_result,)))
        return AsyncConnection(self._context.__enter__(), self.executor)
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.executor.execute(alchemy_incendiary(self._context.__exit__, (exc_type, exc_val, exc_tb),))


class ConnectionCM:
    __slots__ = ('result', 'task',)
    def __init__(self, task):
        self.task = task
        self.result = None
        
    def __await__(self):
        return self.task.__await__()
    
    async def __aenter__(self):
        self.result = await self.task
        return self.result
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.result.close()


class TransactionCM(ConnectionCM):
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None and self.result._transaction.is_active:
            try:
                await self.result.commit()
            except:
                with util.safe_reraise():
                    await self.result.rollback()
        else:
            await self.result.rollback()

KOKORO_STRATEGY = 'KOKORO'

class KOKOROEngineStrategy(DefaultEngineStrategy):
    name = KOKORO_STRATEGY
    engine_cls = KOKOROEngine

KOKOROEngineStrategy()




__all__ = ('ACTIVITY_UNKNOWN', 'ActivityUnknown')

from .activity_base import ActivityBase

class ActivityUnknown(ActivityBase):
    """
    Represents if a user has no activity set. This activity type is not a valid Discord activity.
    
    ``activity_unknown`` is a singleton with type value of `127`.
    
    Class Attributes
    ----------------
    created_at : `datetime`
        When the activity was created. Defaults to Discord epoch.
    color : ``Color`` = `Color(0)
        The color of the activity.
    name : `str` = `'Unknown'`
        The activity's name. Subclasses might overwrite it as member descriptor.
    id : `int` = `0`
        The activity's id. Subclasses might overwrite it as member descriptor.
    type : `int` = `127`
        The activity's type value.
    """
    __slots__ = ()
    
    def __repr__(self):
        """Returns the activity's representation."""
        return f'<{self.__class__.__name__}>'

ACTIVITY_UNKNOWN = object.__new__(ActivityUnknown)



from .attachment import *
from .cross_mention import *
from .flags import *
from .message import *
from .message_application import *
from .message_interaction import *
from .message_reference import *
from .message_repr import *
from .preinstanced import *
from .utils import *

__all__ = (
    *attachment.__all__,
    *cross_mention.__all__,
    *flags.__all__,
    *message.__all__,
    *message_application.__all__,
    *message_interaction.__all__,
    *message_reference.__all__,
    *message_repr.__all__,
    *preinstanced.__all__,
    *utils.__all__,
)



__all__ = ('CHANNEL_TYPE_MAP', 'cr_pg_channel_object', 'create_partial_channel_from_data',
    'create_partial_channel_from_id')

from datetime import datetime
import warnings

from ...backend.export import export, include

from ..bases import maybe_snowflake
from ..core import CHANNELS
from ..utils import datetime_to_timestamp
from ..permission import PermissionOverwrite

from .channel_guild_text import ChannelText
from .channel_private import ChannelPrivate, ChannelGroup
from .channel_guild_voice import ChannelVoice, ChannelStage, ChannelVoiceBase
from .channel_guild_category import ChannelCategory
from .channel_guild_store import ChannelStore
from .channel_thread import ChannelThread
from .channel_guild_undefined import ChannelGuildUndefined
from .preinstanced import VideoQualityMode
from .channel_base import ChannelBase
from .channel_guild_base import ChannelGuildBase
from .channel_thread import AUTO_ARCHIVE_OPTIONS
from .channel_guild_directory import ChannelDirectory
from . import channel_types as CHANNEL_TYPES

VoiceRegion = include('VoiceRegion')
Guild = include('Guild')

CHANNEL_TYPE_MAP = {
    CHANNEL_TYPES.guild_text: ChannelText,
    CHANNEL_TYPES.private: ChannelPrivate,
    CHANNEL_TYPES.guild_voice: ChannelVoice,
    CHANNEL_TYPES.private_group: ChannelGroup,
    CHANNEL_TYPES.guild_category: ChannelCategory,
    CHANNEL_TYPES.guild_announcements: ChannelText,
    CHANNEL_TYPES.guild_store: ChannelStore,
    CHANNEL_TYPES.guild_thread_announcements: ChannelThread,
    CHANNEL_TYPES.guild_thread_public: ChannelThread,
    CHANNEL_TYPES.guild_thread_private: ChannelThread,
    CHANNEL_TYPES.guild_stage: ChannelStage,
    CHANNEL_TYPES.guild_directory: ChannelDirectory,
}

export(CHANNEL_TYPE_MAP, 'CHANNEL_TYPE_MAP')


def create_partial_channel_from_data(data, guild_id):
    """
    Creates a partial channel from partial channel data.
    
    Parameters
    ----------
    data : `None` or `dict` of (`str`, `Any`) items
        Partial channel data received from Discord.
    guild_id : `int`
        The channel's guild's identifier.
    
    Returns
    -------
    channel : `None` or ``ChannelBase`` instance
        The created partial channel, or `None`, if no data was received.
    """
    if (data is None) or (not data):
        return None
    
    channel_id = int(data['id'])
    try:
        return CHANNELS[channel_id]
    except KeyError:
        pass
    
    channel = CHANNEL_TYPE_MAP.get(data['type'], ChannelGuildUndefined)._from_partial_data(data, channel_id, guild_id)
    CHANNELS[channel_id] = channel
    
    return channel


@export
def create_partial_channel_from_id(channel_id, channel_type, guild_id):
    """
    Creates a new partial channel from the given identifier.
    
    Parameters
    ----------
    channel_id : `int`
        The channel's identifier.
    channel_type : `int`
        The channel's type identifier.
    guild_id : `int`
        A guild's identifier of the created channel.
    """
    try:
        return CHANNELS[channel_id]
    except KeyError:
        pass
    
    channel_type = CHANNEL_TYPE_MAP.get(channel_type, ChannelGuildUndefined)
    channel = channel_type._create_empty(channel_id, channel_type, guild_id)
    CHANNELS[channel_id] = channel
    
    return channel


def cr_pg_channel_object(name, type_, *, permission_overwrites=None, topic=None, nsfw=None, slowmode=None, bitrate=None,
        user_limit=None, region=None, video_quality_mode=None, archived=None, archived_at=None,
        auto_archive_after=None, open_=None, default_auto_archive_after=None, parent=None, guild=None, overwrites=None):
    """
    Creates a json serializable object representing a ``GuildChannelBase`` instance.
    
    Parameters
    ----------
    name : `str`
        The name of the channel. Can be between `1` and `100` characters.
    type_ : `int` or ``ChannelGuildBase`` subclass
        The channel's type.
    permission_overwrites : `list` of ``cr_p_permission_overwrite_object`` returns, Optional (Keyword only)
        A list of permission overwrites of the channel. The list should contain json serializable permission
        overwrites made by the ``cr_p_permission_overwrite_object`` function.
    topic : `str`, Optional (Keyword only)
        The channel's topic.
    nsfw : `bool`, Optional (Keyword only)
        Whether the channel is marked as nsfw.
    slowmode : int`, Optional (Keyword only)
        The channel's slowmode value.
    bitrate : `int`, Optional (Keyword only)
        The channel's bitrate.
    user_limit : `int`, Optional (Keyword only)
        The channel's user limit.
    region : `None`, ``VoiceRegion`` or `str`, Optional (Keyword only)
        The channel's voice region.
    video_quality_mode : `None`, ``VideoQualityMode`` or `int`, Optional (Keyword only)
        The channel's video quality mode.
    archived : `None` or `bool`, Optional (Keyword only)
        Whether the thread channel is archived.
    archived_at : `None` or `datetime`, Optional (Keyword only)
        When the thread's archive status was last changed.
    auto_archive_after : `None` or `int`, Optional (Keyword only)
        Duration in minutes to automatically archive the thread after recent activity. Can be one of: `3600`, `86400`,
        `259200`, `604800`.
    open_ : `bool`, Optional (Keyword only)
        Whether the thread channel is open.
    default_auto_archive_after : `None` or `int`
        The default duration (in seconds) for newly created threads to automatically archive the themselves. Can be
        one of: `3600`, `86400`, `259200`, `604800`.
    parent : `None`, ``ChannelCategory`` or `int`, Optional (Keyword only)
        The channel's parent. If the parent is under a guild, leave it empty.
    category : `None`, ``ChannelCategory`` or `int`, Optional (Keyword only)
        Deprecated, please use `parent` parameter instead.
    guild : `None` or ``Guild``, Optional (Keyword only)
        Reference guild used for validation purposes. Defaults to `None`.
    
    Returns
    -------
    channel_data : `dict` of (`str`, `Any`) items
    
    Raises
    ------
    TypeError
        - If `type_` was not passed as `int` or as ``ChannelGuildBase`` instance.
        - If `parent` was not given as `None`, ``ChannelCategory`` or `int` instance.
        - If `region` was not given either as `None`, `str` nor ``VoiceRegion`` instance.
        - If `video_quality_mode` was not given neither as `None`, `VideoQualityMode`` nor as `int` instance.
    AssertionError
        - if `guild` is given, but not as `None` nor ``Guild`` instance.
        - If `type_` was given as `int`, and is less than `0`.
        - If `type_` was given as `int` and exceeds the defined channel type limit.
        - If `name` was not given as `str` instance.
        - If `name`'s length is under `1` or over `100`.
        - If `permission_overwrites` was not given as `None`, neither as `list` of `dict`-s.
        - If `topic` was not given as `str` instance.
        - If `topic`'s length is over `1024` or `120` depending on channel type.
        - If `topic` was given, but the respective channel type is not ``ChannelText`` nor ``ChannelStage``.
        - If `nsfw` was given meanwhile the respective channel type is not ``ChannelText`` or ``ChannelStore``.
        - If `nsfw` was not given as `bool`.
        - If `slowmode` was given, but the respective channel type is not ``ChannelText`` or ``ChannelThread``.
        - If `slowmode` was not given as `int` instance.
        - If `slowmode` was given, but it's value is less than `0` or greater than `21600`.
        - If `bitrate` was given, but the respective channel type is not ``ChannelVoiceBase``.
        - If `bitrate` was not given as `int` instance.
        - If `bitrate`'s value is out of the expected range.
        - If `user_limit` was given, but the respective channel type is not ``ChannelVoiceBase``.
        - If `user_limit` was not given as `int` instance.
        - If `user_limit`' was given, but is out of the expected [0:99] range.
        - If `parent` was given, but the respective channel type cannot be put under other categories.
        - If `region` was given, but the respective channel type is not ``ChannelVoiceBase``.
        - If `video_quality_mode` was given, but the respective channel is not ``ChannelVoice`` instance.
        - If `archived` was given meanwhile the respective channel type is not ``ChannelThread``.
        - If `archived` was given, but not as `None` or `bool` instance.
        - If `archived_at` was given meanwhile the respective channel type is not ``ChannelThread``.
        - If `archived_at` was given, but not as `None` or `datetime` instance.
        - If `auto_archive_after` was given, but the respective channel's type is not ``ChannelThread``.
        - If `auto_archive_after` was given, but not as `None` or `int` instance.
        - If `auto_archive_after` is not any of the expected values.
        - If `open_` was given meanwhile the respective channel type is not ``ChannelThread``.
        - If `open_` was given, but not as `None` or `bool` instance.
        - If `default_auto_archive_after` was given as non `None`, but the respective channel type is not `ChannelText`.
        - If `default_auto_archive_after` was not given neither as `None` or `int` instance.
        - If `default_auto_archive_after` is not any of the expected values.
    """
    if (overwrites is not None):
        warnings.warn(
            f'`cr_pg_channel_object`\'s `overwrites` parameter is deprecated, and will be removed in 2021 November. '
            f'Please use `permission_overwrites` instead.',
            FutureWarning)
        
        permission_overwrites = overwrites
    
    if __debug__:
        if (guild is not None) and (not isinstance(guild, Guild)):
            raise AssertionError('`guild` is given, but not as `None` nor `Guild` instance, got '
                f'{guild.__class__.__name__}.')
    
    if isinstance(type_, int):
        if __debug__:
            if type_ < 0:
                raise AssertionError(f'`type_` cannot be negative value, got `{type_!r}`.')
            if type_ not in CHANNEL_TYPE_MAP.keys():
                raise AssertionError(f'`type_` is not in an of the existing channel types: '
                    f'{set(CHANNEL_TYPE_MAP.keys())!r}, got `{type_}`.')
        
        channel_type = CHANNEL_TYPE_MAP.get(type_, ChannelGuildUndefined)
        channel_type_value = type_
    
    elif issubclass(type_, ChannelBase):
        channel_type = type_
        channel_type_value = type_.INTERCHANGE[0]
    else:
        raise TypeError(f'The given `type_` is not, neither refers to a channel a type, got {type_!r}.')
    
    if not issubclass(channel_type, ChannelGuildBase):
        raise TypeError(f'`type_` not refers to a `{ChannelGuildBase.__name__}` instance, but to '
            f'{channel_type.__name__}. Got {type_!r}.')
    
    
    if __debug__:
        if not isinstance(name, str):
            raise AssertionError(f'`name` can be given as `str` instance, got {name.__class__.__name__}.')
        
        name_length = len(name)
        
        if name_length < 1 or name_length > 100:
            raise AssertionError(f'`name` length can be in range [1:100], got {name_length}; {name!r}.')
    
    channel_data = {
        'name': name,
        'type': channel_type_value,
    }
    
    if not issubclass(channel_type, ChannelThread):
        if permission_overwrites is None:
            permission_overwrites = []
        else:
            if __debug__:
                if not isinstance(permission_overwrites, list):
                    raise AssertionError(f'`permission_overwrites` can be given as `None` or `list` of '
                        f'`cr_p_permission_overwrite_object` returns, got {permission_overwrites.__class__.__name__}')
                
                for index, element in enumerate(permission_overwrites):
                    if not isinstance(element, dict):
                        raise AssertionError(f'`permission_overwrites`\'s element {index} should be `dict` instance, '
                            f'but got {element.__class__.__name__}')
        
        channel_data['permission_overwrites'] = permission_overwrites
    
    
    if (topic is not None):
        if __debug__:
            if not issubclass(channel_type, (ChannelText, ChannelStage)):
                raise AssertionError(f'`topic` is a valid parameter only for `{ChannelText.__name__}` and for '
                    f'{ChannelStage.__name__} instances, got {channel_type.__name__}.')
            
            if not isinstance(topic, str):
                raise AssertionError(f'`topic` can be given as `str` instance, got {topic.__class__.__name__}.')
            
            if issubclass(channel_type, ChannelText):
                topic_length_limit = 1024
            else:
                topic_length_limit = 120
            
            topic_length = len(topic)
            if topic_length > topic_length_limit:
                raise AssertionError(f'`topic` length can be in range [0:{topic_length_limit}], got {topic_length}; '
                    f'{topic!r}.')
        
        channel_data['topic'] = topic
    
    
    if (nsfw is not None):
        if __debug__:
            if not issubclass(channel_type, (ChannelText, ChannelStore)):
                raise AssertionError(f'`nsfw` is a valid parameter only for `{ChannelText.__name__}` and '
                    f'`{ChannelStore.__name__}` instances, but got {channel_type.__name__}.')
            
            if not isinstance(nsfw, bool):
                raise AssertionError(f'`nsfw` can be given as `bool` instance, got {nsfw.__class__.__name__}.')
        
        channel_data['nsfw'] = nsfw
    
    
    if (slowmode is not None):
        if __debug__:
            if not issubclass(channel_type, (ChannelText, ChannelThread)):
                raise AssertionError(f'`slowmode` is a valid parameter only for `{ChannelText.__name__}` and for '
                    f'`{ChannelThread.__name__}` instances, but got {channel_type.__name__}.')
            
            if not isinstance(slowmode, int):
                raise AssertionError('`slowmode` can be given as `int` instance, got '
                    f'{slowmode.__class__.__name__}.')
            
            if slowmode < 0 or slowmode > 21600:
                raise AssertionError(f'`slowmode` can be in range [0:21600], got: {slowmode!r}.')
        
        channel_data['rate_limit_per_user'] = slowmode
    
    
    if (bitrate is not None):
        if __debug__:
            if not issubclass(channel_type, ChannelVoiceBase):
                raise AssertionError(f'`bitrate` is a valid parameter only for `{ChannelVoiceBase.__name__}` instances, '
                    f'but got {channel_type.__name__}.')
                
            if not isinstance(bitrate, int):
                raise AssertionError('`bitrate` can be given as `int` instance, got '
                    f'{bitrate.__class__.__name__}.')
            
            # Get max bitrate
            if guild is None:
                bitrate_limit = 384000
            else:
                bitrate_limit = guild.bitrate_limit
            
            if bitrate < 8000 or bitrate > bitrate_limit:
                raise AssertionError(f'`bitrate` is out of the expected [8000:{bitrate_limit}] range, got {bitrate!r}.')
        
        channel_data['bitrate'] = bitrate
    
    
    if (user_limit is not None):
        if __debug__:
            if not issubclass(channel_type, ChannelVoiceBase):
                raise AssertionError(f'`user_limit` is a valid parameter only for `{ChannelVoiceBase.__name__}` '
                    f'instances, but got {channel_type.__name__}.')
            
            if user_limit < 0 or user_limit > 99:
                raise AssertionError('`user_limit`\'s value is out of the expected [0:99] range, got '
                    f'{user_limit!r}.')
        
        channel_data['user_limit'] = user_limit
    
    
    if (region is not None):
        if __debug__:
            if not issubclass(channel_type, ChannelVoiceBase):
                raise AssertionError(f'`region` is a valid parameter only for `{ChannelVoiceBase.__name__}` '
                    f'instances, but got {channel_type.__name__}.')
        
        if isinstance(region, VoiceRegion):
            region_value = region.value
        elif isinstance(region, str):
            region_value = region
        else:
            raise TypeError(f'`region` can be given either as `None`, `str` or as `{VoiceRegion.__name__}` instance, '
                f'{region.__class__.__name__}.')
        
        channel_data['rtc_region'] = region_value
    
    
    if (video_quality_mode is not None):
        if __debug__:
            if not issubclass(channel_type, ChannelVoice):
                raise AssertionError(f'`video_quality_mode` is a valid parameter only for `{ChannelVoice.__name__}` '
                    f'instances, but got {channel_type.__name__}.')
        
        if isinstance(video_quality_mode, VideoQualityMode):
            video_quality_mode_value = video_quality_mode.value
        elif isinstance(video_quality_mode, int):
            video_quality_mode_value = video_quality_mode
        else:
            raise TypeError(f'`video_quality_mode` can be given either as `None`, `str` or as '
                f'`{VideoQualityMode.__name__}` instance, {video_quality_mode.__class__.__name__}.')
        
        channel_data['video_quality_mode'] = video_quality_mode_value
    
    
    if (archived is not None):
        if __debug__:
            if not issubclass(channel_type, ChannelThread):
                raise AssertionError(f'`archived` is a valid parameter only for `{ChannelThread.__name__}` '
                    f'instances, but got {channel_type.__name__}.')
            
            if not isinstance(archived, bool):
                raise AssertionError(f'`archived` can be given as `None` or as `bool` instance, got '
                    f'{archived.__class__.__name__}.')
        
        channel_data['archived'] = archived
    
    
    if (archived_at is not None):
        if __debug__:
            if not issubclass(channel_type, ChannelThread):
                raise AssertionError(f'`archived_at` is a valid parameter only for `{ChannelThread.__name__}` '
                    f'instances, but got {channel_type.__name__}.')
            
            if not isinstance(archived_at, datetime):
                raise AssertionError(f'`archived_at` can be given as `None` or as `datetime` instance, got '
                    f'{archived_at.__class__.__name__}.')
        
        channel_data['archive_timestamp'] =  datetime_to_timestamp(archived_at)
    
    
    if (auto_archive_after is not None):
        if __debug__:
            if not issubclass(channel_type, ChannelThread):
                raise AssertionError(f'`auto_archive_after` is a valid parameter only for `{ChannelThread.__name__}` '
                    f'instances, but got {channel_type.__name__}.')
            
            if not isinstance(auto_archive_after, int):
                raise AssertionError(f'`auto_archive_after` can be given as `None` or as `datetime` instance, got '
                    f'{auto_archive_after.__class__.__name__}.')
            
            if auto_archive_after not in AUTO_ARCHIVE_OPTIONS:
                raise AssertionError(f'`auto_archive_after` can be any of: '
                    f'{AUTO_ARCHIVE_OPTIONS}, got {auto_archive_after}.')
        
        channel_data['auto_archive_duration'] = auto_archive_after//60
    
    
    if (default_auto_archive_after is not None):
        if __debug__:
            if not issubclass(channel_type, ChannelText):
                raise AssertionError(f'`default_auto_archive_after` is a valid parameter only for `{ChannelText.__name__}` '
                    f'instances, but got {channel_type.__name__}.')
            
            if not isinstance(default_auto_archive_after, int):
                raise AssertionError(f'`default_auto_archive_after` can be given as `None` or as `datetime` instance, got '
                    f'{default_auto_archive_after.__class__.__name__}.')
            
            if default_auto_archive_after not in AUTO_ARCHIVE_OPTIONS:
                raise AssertionError(f'`default_auto_archive_after` can be any of: '
                    f'{AUTO_ARCHIVE_OPTIONS}, got {default_auto_archive_after}.')
        
        channel_data['default_auto_archive_duration'] = default_auto_archive_after//60
    
    
    if (open_ is not None):
        if __debug__:
            if not issubclass(channel_type, ChannelThread):
                raise AssertionError(f'`open_` is a valid parameter only for `{ChannelThread.__name__}` '
                    f'instances, but got {channel_type.__name__}.')
            
            if not isinstance(open_, bool):
                raise AssertionError(f'`open_` can be given as `None` or as `bool` instance, got '
                    f'{open_.__class__.__name__}.')
        
        channel_data['locked'] = not open_
    
    
    if parent is None:
        parent_id = 0
    elif isinstance(parent, ChannelCategory):
        parent_id = parent.id
    else:
        parent_id = maybe_snowflake(parent)
        if parent_id is None:
            raise TypeError(f'`parent` can be given as `{ChannelCategory.__name__}`, `{Guild.__name__}` or `int` '
                f'instance, got {parent.__class__.__name__}.')
    
    if parent_id:
        if __debug__:
            if issubclass(channel_type, ChannelCategory):
                raise AssertionError(f'`parent` was given, but the respective channel type is '
                    f'{channel_type.__name__}, which cannot be put under other categories.')
        
        channel_data['parent_id'] = parent_id
    
    return channel_data


@export
def parse_permission_overwrites(data):
    """
    Parses the permission overwrites from the given data and returns them.
    
    Parameters
    ----------
    data : `list` of (`dict` of (`str`, `Any`) items) elements
        A list of permission overwrites' data.
    
    Returns
    -------
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
    """
    permission_overwrites = {}
    
    permission_overwrites_datas = data.get('permission_overwrites', None)
    if (permission_overwrites_datas is not None) and permission_overwrites_datas:
        for permission_overwrite_data in permission_overwrites_datas:
            permission_overwrite = PermissionOverwrite(permission_overwrite_data)
            permission_overwrites[permission_overwrite.target_id] = permission_overwrite
    
    return permission_overwrites



__all__ = ('ExtensionError', )


class ExtensionError(Exception):
    """
    An exception raised by the ``ExtensionLoader``, if loading, reloading or unloading an extension fails with any
    reason.
    
    Attributes
    ----------
    _message : `str` or `list` of `str`
        The error's message.
    """
    
    def __init__(self, message):
        """
        Creates a new extension error.
        
        Parameters
        ----------
        message : `str` or `list` of `str`
            The error's message.
        """
        self._message = message
    
    @property
    def message(self):
        """
        Returns the extension error's message.
        
        If the extension error contains more message,s connects them.
        
        Returns
        -------
        message : ``Message``
        """
        message = self._message
        if isinstance(message, str):
            return message
        
        return '\n\n'.join(message)
    
    @property
    def messages(self):
        """
        Returns a list containing the exception error's messages.
        
        Returns
        -------
        messages : `list` of `str`
        """
        message = self._message
        if isinstance(message, str):
           return [message]
        
        return message
    
    def __len__(self):
        """Returns the amount of messages, what the extension error contains."""
        message = self._message
        if isinstance(message, str):
            return 1
        
        return len(message)
    
    def __repr__(self):
        """Returns the exception error's representation."""
        return f'{self.__class__.__name__} ({len(self)}):\n{self.message}\n'
    
    __str__ = __repr__


class DoNotLoadExtension(BaseException):
    """
    Raised to stop an extension loaded without error.
    """
